# -*- coding: utf-8 -*-

##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

"""
Elementary mathematical functions and conditional masks for escript.

This module provides mathematical functions (sin, cos, exp, sqrt, etc.) and
conditional mask functions (wherePositive, whereNegative, etc.) that work
with escript Data objects, numpy arrays, and symbolic expressions.
"""


__copyright__ = """Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__ = """Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__ = "https://github.com/LutzGross/esys-escript.github.io"

import math
import cmath
import numpy

from . import escriptcpp as escore
from .start import HAVE_SYMBOLS

if HAVE_SYMBOLS:
    from . import symboliccore as sym
else:
    class sym:
        Symbol = type(memoryview(b'1'))  # This type should never be passed

# Machine epsilon for tolerance calculations
EPSILON = escore.getMachinePrecision()

__all__ = [
    'log10', 'wherePositive', 'whereNegative', 'whereNonNegative',
    'whereNonPositive', 'whereZero', 'whereNonZero', 'Abs', 'erf',
    'sin', 'cos', 'tan', 'asin', 'acos', 'atan', 'atan2',
    'sinh', 'cosh', 'tanh', 'asinh', 'acosh', 'atanh',
    'exp', 'sqrt', 'log', 'sign', 'minval', 'maxval'
]


def log10(arg):
    """
    Returns base-10 logarithm of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.log10(arg)
    elif isinstance(arg, escore.Data):
        return arg._log10()
    elif isinstance(arg, complex):
        return cmath.log10(arg)
    elif isinstance(arg, float) or isinstance(arg, int):
        return math.log10(arg)
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.log10)
    else:
        raise TypeError("log10: Unknown argument type.")


def wherePositive(arg):
    """
    Returns mask of positive values of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``.
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        if arg.dtype.kind == 'c':
            raise TypeError("wherePositive: operation not supported for complex")
        out = numpy.greater(arg, numpy.zeros(arg.shape, numpy.double)) * 1.
        return out
    elif isinstance(arg, escore.Data):
        return arg._wherePositive()
    elif isinstance(arg, float):
        if arg > 0:
            return 1.
        else:
            return 0.
    elif isinstance(arg, int):
        if arg > 0:
            return 1.
        else:
            return 0.
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.wherePositive)
    elif isinstance(arg, complex):
        raise TypeError("wherePositive: operation not supported for complex")
    else:
        raise TypeError("wherePositive: Unknown argument type.")


def whereNegative(arg):
    """
    Returns mask of negative values of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        if arg.dtype.kind == 'c':
            raise TypeError("whereNegative: operation not supported for complex")
        out = numpy.less(arg, numpy.zeros(arg.shape, numpy.double)) * 1.
        return out
    elif isinstance(arg, escore.Data):
        return arg._whereNegative()
    elif isinstance(arg, float):
        if arg < 0:
            return 1.
        else:
            return 0.
    elif isinstance(arg, int):
        if arg < 0:
            return 1.
        else:
            return 0.
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.whereNegative)
    elif isinstance(arg, complex):
        raise TypeError("whereNegative: operation not supported for complex")
    else:
        raise TypeError("whereNegative: Unknown argument type.")


def whereNonNegative(arg):
    """
    Returns mask of non-negative values of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        if arg.dtype.kind == 'c':
            raise TypeError("whereNonNegative: operation not supported for complex")
        out = numpy.greater_equal(arg, numpy.zeros(arg.shape, numpy.double)) * 1.
        return out
    elif isinstance(arg, escore.Data):
        return arg._whereNonNegative()
    elif isinstance(arg, float):
        if arg < 0:
            return 0.
        else:
            return 1.
    elif isinstance(arg, int):
        if arg < 0:
            return 0.
        else:
            return 1.
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.whereNonNegative)
    elif isinstance(arg, complex):
        raise TypeError("whereNonNegative: operation not supported for complex")
    else:
        raise TypeError("whereNonNegative: Unknown argument type.")


def whereNonPositive(arg):
    """
    Returns mask of non-positive values of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        if arg.dtype.kind == 'c':
            raise TypeError("whereNonPositive: operation not supported for complex")
        out = numpy.less_equal(arg, numpy.zeros(arg.shape, numpy.double)) * 1.
        return out
    elif isinstance(arg, escore.Data):
        return arg._whereNonPositive()
    elif isinstance(arg, float):
        if arg > 0:
            return 0.
        else:
            return 1.
    elif isinstance(arg, int):
        if arg > 0:
            return 0.
        else:
            return 1.
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.whereNonPositive)
    elif isinstance(arg, complex):
        raise TypeError("whereNonPositive: operation not supported for complex")
    else:
        raise TypeError("whereNonPositive: Unknown argument type.")


def whereZero(arg, tol=None, rtol=math.sqrt(EPSILON)):
    """
    Returns mask of zero entries of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :param tol: absolute tolerance. Values with absolute value less than tol are accepted
                as zero. If ``tol`` is not present ``rtol * Lsup(arg)`` is used.
    :type tol: ``float``
    :param rtol: relative tolerance used to define the absolute tolerance if ``tol`` is not present.
    :type rtol: non-negative ``float``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise ValueError: if ``rtol`` is negative.
    :raise TypeError: if the type of the argument is not expected
    """
    if tol is None and not isinstance(arg, sym.Symbol):
        if rtol < 0:
            raise ValueError("rtol must be non-negative.")
        # Import Lsup locally to avoid circular import
        from . import util
        tol = util.Lsup(arg) * rtol
    if isinstance(arg, numpy.ndarray):
        out = numpy.less_equal(abs(arg) - tol, numpy.zeros(arg.shape, numpy.double)) * 1.
        if isinstance(out, float):
            out = numpy.array(out, dtype=numpy.double)
        return out
    elif isinstance(arg, escore.Data):
        return arg._whereZero(tol)
    elif isinstance(arg, float) or isinstance(arg, complex) or isinstance(arg, int):
        if abs(arg) <= tol:
            return 1.
        else:
            return 0.
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.whereZero)
    else:
        raise TypeError("whereZero: Unknown argument type.")


def whereNonZero(arg, tol=0.):
    """
    Returns mask of values different from zero of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :param tol: absolute tolerance. Values with absolute value less than or equal
                to tol are considered zero.
    :type tol: ``float``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        out = numpy.greater(abs(arg) - tol, numpy.zeros(arg.shape, numpy.double)) * 1.
        if isinstance(out, float):
            out = numpy.array(out, dtype=numpy.double)
        return out
    elif isinstance(arg, escore.Data):
        return arg._whereNonZero(tol)
    elif isinstance(arg, float) or isinstance(arg, complex) or isinstance(arg, int):
        if abs(arg) > tol:
            return 1.
        else:
            return 0.
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.whereNonZero)
    else:
        raise TypeError("whereNonZero: Unknown argument type.")


def Abs(arg):
    """
    Returns the absolute value of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``.
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.abs)
    else:
        return abs(arg)


def erf(arg):
    """
    Returns the error function *erf* of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``.
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, escore.Data):
        return arg._erf()
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.erf)
    else:
        raise TypeError("erf: Unknown argument type.")


def sin(arg):
    """
    Returns sine of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``.
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.sin(arg)
    elif isinstance(arg, escore.Data):
        return arg._sin()
    elif isinstance(arg, complex):
        return cmath.sin(arg)
    elif isinstance(arg, float) or isinstance(arg, int):
        return math.sin(arg)
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.sin)
    else:
        raise TypeError("sin: Unknown argument type.")


def cos(arg):
    """
    Returns cosine of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.cos(arg)
    elif isinstance(arg, escore.Data):
        return arg._cos()
    elif isinstance(arg, complex):
        return cmath.cos(arg)
    elif isinstance(arg, float) or isinstance(arg, int):
        return math.cos(arg)
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.cos)
    else:
        raise TypeError("cos: Unknown argument type.")


def tan(arg):
    """
    Returns tangent of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.tan(arg)
    elif isinstance(arg, escore.Data):
        return arg._tan()
    elif isinstance(arg, complex):
        return cmath.tan(arg)
    elif isinstance(arg, float) or isinstance(arg, int):
        return math.tan(arg)
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.tan)
    else:
        raise TypeError("tan: Unknown argument type.")


def asin(arg):
    """
    Returns the inverse sine of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.arcsin(arg)
    elif isinstance(arg, escore.Data):
        return arg._asin()
    elif isinstance(arg, complex):
        return cmath.asin(arg)
    elif isinstance(arg, float) or isinstance(arg, int):
        return math.asin(arg)
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.asin)
    else:
        raise TypeError("asin: Unknown argument type.")


def acos(arg):
    """
    Returns the inverse cosine of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.arccos(arg)
    elif isinstance(arg, escore.Data):
        return arg._acos()
    elif isinstance(arg, complex):
        return cmath.acos(arg)
    elif isinstance(arg, float) or isinstance(arg, int):
        return math.acos(arg)
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.acos)
    else:
        raise TypeError("acos: Unknown argument type.")


def atan(arg):
    """
    Returns inverse tangent of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.arctan(arg)
    elif isinstance(arg, escore.Data):
        return arg._atan()
    elif isinstance(arg, complex):
        return cmath.atan(arg)
    elif isinstance(arg, float) or isinstance(arg, int):
        return math.atan(arg)
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.atan)
    else:
        raise TypeError("atan: Unknown argument type.")


def atan2(arg0, arg1):
    """
    Returns inverse tangent of ``arg0/arg1``, handling quadrant correctly.

    :param arg0: numerator argument (y coordinate)
    :type arg0: ``float``, `escript.Data`, ``numpy.ndarray``
    :param arg1: denominator argument (x coordinate)
    :type arg1: ``float``, `escript.Data`, ``numpy.ndarray``
    :return: angle in radians between -pi and pi
    :rtype: ``float``, `escript.Data`, ``numpy.ndarray`` depending on input types
    """
    m = whereZero(arg1, rtol=EPSILON)
    m2 = whereNegative(arg1 * arg0)
    s = atan(arg0 / (arg1 + m)) * (1 - m) + (numpy.pi / 2) * (1 - 2 * m2) * m
    s += (wherePositive(arg0) * whereNegative(arg1) - whereNegative(arg0) * whereNegative(arg1)) * numpy.pi
    return s


def sinh(arg):
    """
    Returns the hyperbolic sine of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.sinh(arg)
    elif isinstance(arg, escore.Data):
        return arg._sinh()
    elif isinstance(arg, complex):
        return cmath.sinh(arg)
    elif isinstance(arg, float) or isinstance(arg, int):
        return math.sinh(arg)
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.sinh)
    else:
        raise TypeError("sinh: Unknown argument type.")


def cosh(arg):
    """
    Returns the hyperbolic cosine of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.cosh(arg)
    elif isinstance(arg, escore.Data):
        return arg._cosh()
    elif isinstance(arg, complex):
        return cmath.cosh(arg)
    elif isinstance(arg, float) or isinstance(arg, int):
        return math.cosh(arg)
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.cosh)
    else:
        raise TypeError("cosh: Unknown argument type.")


def tanh(arg):
    """
    Returns the hyperbolic tangent of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.tanh(arg)
    elif isinstance(arg, escore.Data):
        return arg._tanh()
    elif isinstance(arg, complex):
        return cmath.tanh(arg)
    elif isinstance(arg, float) or isinstance(arg, int):
        return math.tanh(arg)
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.tanh)
    else:
        raise TypeError("tanh: Unknown argument type.")


def asinh(arg):
    """
    Returns the inverse hyperbolic sine of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.arcsinh(arg)
    elif isinstance(arg, escore.Data):
        return arg._asinh()
    elif isinstance(arg, complex):
        return numpy.arcsinh(complex(arg))
    elif isinstance(arg, float) or isinstance(arg, int):
        return numpy.arcsinh(float(arg))
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.asinh)
    else:
        raise TypeError("asinh: Unknown argument type.")


def acosh(arg):
    """
    Returns the inverse hyperbolic cosine of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.arccosh(arg)
    elif isinstance(arg, escore.Data):
        return arg._acosh()
    elif isinstance(arg, complex):
        return numpy.arccosh(complex(arg))
    elif isinstance(arg, float) or isinstance(arg, int):
        return numpy.arccosh(float(arg))
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.acosh)
    else:
        raise TypeError("acosh: Unknown argument type.")


def atanh(arg):
    """
    Returns the inverse hyperbolic tangent of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.arctanh(arg)
    elif isinstance(arg, escore.Data):
        return arg._atanh()
    elif isinstance(arg, complex):
        return numpy.arctanh(complex(arg))
    elif isinstance(arg, float) or isinstance(arg, int):
        return numpy.arctanh(float(arg))
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.atanh)
    else:
        raise TypeError("atanh: Unknown argument type.")


def exp(arg):
    """
    Returns *e* to the power of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``.
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of arg
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.exp(arg)
    elif isinstance(arg, escore.Data):
        return arg._exp()
    elif isinstance(arg, complex):
        return cmath.exp(arg)
    elif isinstance(arg, float) or isinstance(arg, int):
        return math.exp(arg)
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.exp)
    else:
        raise TypeError("exp: Unknown argument type.")


def sqrt(arg):
    """
    Returns the square root of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
            depending on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.sqrt(arg)
    elif isinstance(arg, escore.Data):
        return arg._sqrt()
    elif isinstance(arg, complex):
        return cmath.sqrt(arg)
    elif isinstance(arg, float) or isinstance(arg, int):
        return math.sqrt(arg)
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.sqrt)
    else:
        raise TypeError("sqrt: Unknown argument type.")


def log(arg):
    """
    Returns the natural logarithm of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``.
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        return numpy.log(arg)
    elif isinstance(arg, escore.Data):
        return arg._log()
    elif isinstance(arg, complex):
        return cmath.log(arg)
    elif isinstance(arg, float) or isinstance(arg, int):
        return math.log(arg)
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.log)
    else:
        raise TypeError("log: Unknown argument type.")


def sign(arg):
    """
    Returns the sign of argument ``arg``.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray`` depending
            on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        if arg.dtype.kind == 'c':
            raise TypeError("sign: operation not supported for complex")
        return wherePositive(arg) - whereNegative(arg)
    elif isinstance(arg, escore.Data):
        return arg._sign()
    elif isinstance(arg, complex):
        raise TypeError("sign: operation not supported for complex")
    elif isinstance(arg, float) or isinstance(arg, int):
        if arg > 0:
            return 1.
        elif arg < 0:
            return -1.
        else:
            return 0.
    elif isinstance(arg, sym.Symbol):
        return arg.applyfunc(sym.symfn.sign)
    else:
        raise TypeError("sign: Unknown argument type.")


def minval(arg):
    """
    Returns the minimum value over all components of ``arg`` at each data point.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol` depending on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        if arg.dtype.kind == 'c':
            raise TypeError("minval: operation not supported for complex")
        if arg.ndim == 0:
            return float(arg)
        else:
            return arg.min()
    elif isinstance(arg, escore.Data):
        return arg._minval()
    elif isinstance(arg, float):
        return arg
    elif isinstance(arg, int):
        return float(arg)
    elif isinstance(arg, sym.Symbol):
        return sym.symfn.minval(arg)
    else:
        raise TypeError("minval: Unknown argument type.")


def maxval(arg):
    """
    Returns the maximum value over all components of ``arg`` at each data point.

    :param arg: argument
    :type arg: ``float``, `escript.Data`, `Symbol`, ``numpy.ndarray``
    :rtype: ``float``, `escript.Data`, `Symbol` depending on the type of ``arg``
    :raise TypeError: if the type of the argument is not expected
    """
    if isinstance(arg, numpy.ndarray):
        if arg.dtype.kind == 'c':
            raise TypeError("maxval: operation not supported for complex")
        if arg.ndim == 0:
            return float(arg)
        else:
            return arg.max()
    elif isinstance(arg, escore.Data):
        return arg._maxval()
    elif isinstance(arg, float):
        return arg
    elif isinstance(arg, int):
        return float(arg)
    elif isinstance(arg, sym.Symbol):
        return sym.symfn.maxval(arg)
    else:
        raise TypeError("maxval: Unknown argument type.")



##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################


# This example demonstrates table interpolation and saving data in CSV format.
# It corresponds to Section "Interpolating Data" in the User Guide.

from esys.escript import InterpolationTable, saveDataCSV, sup, mkDir
import numpy
try:
    from esys.finley import Rectangle, Brick
    HAVE_FINLEY = True
except ImportError:
    HAVE_FINLEY = False

if not HAVE_FINLEY:
    print("Finley module not available")
else:
    mkDir("output")
    n = 4
    r = Rectangle(n, n)
    x = r.getX()   # Data with shape (2,)
    toobig = 100   # RuntimeError if any interpolated value exceeds this

    # ------------------------------------------------------------------ #
    # 1-D interpolation — order-1 (linear)
    # Approximate a sine curve over [0, 1]
    # ------------------------------------------------------------------ #

    sine_table = numpy.array([0, 0.70710678118654746, 1, 0.70710678118654746, 0,
                               -0.70710678118654746, -1, -0.70710678118654746, 0])
    minval, maxval = 0., 1.
    # n-1 intervals between n sample points
    step = (maxval - minval) / (len(sine_table) - 1)

    interp1d = InterpolationTable(sine_table, origin=minval, step=step,
                                  order=1, undef=toobig)
    result = interp1d(x[0])
    saveDataCSV("output/1d.csv", inp=x[0], out=result)

    # ------------------------------------------------------------------ #
    # 1-D interpolation — order-0 (piecewise constant / cell-centred)
    # Each cell i covers [origin + i*step, origin + (i+1)*step)
    # Use extend= to specify total domain width directly
    # ------------------------------------------------------------------ #

    interp1d_o0 = InterpolationTable(sine_table, origin=minval,
                                     extend=(maxval - minval),
                                     order=0, undef=toobig)
    result0 = interp1d_o0(x[0])

    # ------------------------------------------------------------------ #
    # 2-D interpolation — order-1 using "ijk" convention (default)
    # table[ix, iy] = value at (xmin + ix*xstep, ymin + iy*ystep)
    # Table shape: (nx, ny)
    # ------------------------------------------------------------------ #

    nx = ny = 10
    xstep = ystep = (maxval - minval) / (nx - 1)
    xmin = ymin = minval

    # Sine amplitude decreases from 1 at y=0 to 0 at y=maxval
    st = sine_table  # length 9
    # Build a 9 x 3 table: table[ix, iy], shape (9, 3)
    # y covers [0, 0.55*2] in 3 rows (just as in the original example)
    table2 = numpy.array([st, 0.5 * st, 0. * st]).T  # shape (9, 3)
    x2step = step        # step along x (sine)
    y2step = 0.55        # step along y

    interp2d = InterpolationTable(table2, origin=(minval, 0.),
                                  step=(x2step, y2step),
                                  order=1, undef=toobig)
    result2 = interp2d(x)   # x has shape (2,)
    saveDataCSV("output/2d.csv", inp0=x[0], inp1=x[1], out=result2)

    # ------------------------------------------------------------------ #
    # 3-D interpolation — order-1 using "ijk" convention (default)
    # table[ix, iy, iz] = ix + iy*10 + iz*100
    # Table shape: (nx, ny, nz)
    # ------------------------------------------------------------------ #

    b = Brick(n, n, n)
    x3 = b.getX()   # shape (3,)
    toobig3 = 1000000

    nx3 = ny3 = nz3 = 10
    xstep3 = ystep3 = zstep3 = (maxval - minval) / (nx3 - 1)

    table3 = numpy.array([[[ix + iy * 10 + iz * 100
                             for iz in range(nz3)]
                            for iy in range(ny3)]
                           for ix in range(nx3)], dtype=float)

    interp3d = InterpolationTable(table3,
                                  origin=(minval, minval, minval),
                                  step=(xstep3, ystep3, zstep3),
                                  order=1, undef=toobig3)
    result3 = interp3d(x3)



##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# Development 2012-2013 by School of Earth Sciences
# Development from 2014 by Centre for Geoscience Computing (GeoComp)
# Development from 2019 by School of Earth and Environmental Sciences
#
##############################################################################

import os, platform, re, setuptools, setuptools.command.build_ext, setuptools.command.install, sys

esys_version = '5.6.2'
python_version = '%d%d' % sys.version_info[:2]
python_version_site = '%d.%d' % sys.version_info[:2]
system, machine = platform.system().lower(), platform.machine().lower()
url = 'https://github.com/esys-escript/esys-escript.github.io/releases/download'

if 'linux' in system and '64' in machine:
    plat = 'Linux64'
    py_lib_ext = 'so'
    url_lib_ext = 'tgz'
elif 'windows' in system and '64' in machine:
    plat = 'Windows64'
    py_lib_ext = 'pyd'
    url_lib_ext = 'zip'
# elif 'darwin' in system:
#     plat = 'MacOSX'

# check we have the required binaries
if plat is None:
    raise Exception('Platform not supported: py{}-{}-{}'.format(python_version, system, machine))

esys_bin = '-'.join([plat, 'py'+python_version, 'lib'])
esys_lib = 'esys_escript_lib'

class esys_install(setuptools.command.install.install):
    ''' install platform specific libraries '''

    def _download(self):
        import requests, tarfile, zipfile
        fname = esys_bin + '.' + url_lib_ext
        _url = '/'.join([url, esys_version, fname])
        if not os.path.exists(fname):
            print('downloading {}...'.format(_url))
            resp = requests.get(_url, allow_redirects=True, verify=False)
            if resp.status_code != requests.codes.ok:
                if resp.status_code == requests.codes.not_found:
                    print('\nescript platform-specific libs not found: ' + fname)
                    print('please see available libs at:\n' + url + '\n')
                resp.raise_for_status()
            with open(fname, 'wb') as f:
                f.write(resp.content)
        print('extracting {}...'.format(fname))
        tar = tarfile.open(fname) if url_lib_ext == 'tgz' else zipfile.ZipFile(fname, 'r')
        tar.extractall()

    def _updateBuildvars(self, path):
        prefix_rep = re.compile('^(prefix=).*$')
        python_rep = re.compile('^(python=).*$')
        inc_path_rep = re.compile('^(.*_inc_path=).*$')
        lib_path_rep = re.compile('^(.*_lib_path=).*$')
        fn_1 = path
        fn_2 = path + '.in'

        with open(fn_1) as f:
            lines = f.readlines()

        for i, line in enumerate(lines):
            line = line.strip()
            m = prefix_rep.match(line)
            if m:
                lines[i] = m.group(1) + '%(prefix)s\n'
            m = python_rep.match(line)
            if m:
                lines[i] = m.group(1) + '%(python)s\n'
            m = inc_path_rep.match(line)
            if m:
                lines[i] = m.group(1) + '%(inc_prefix)s\n'
            m = lib_path_rep.match(line)
            if m:
                lines[i] = m.group(1) + '%(lib_prefix)s\n'

        with open(fn_2, 'w') as f:
            f.writelines(lines)
        
    def run(self):
        self._download()
        if plat.startswith('Windows'):
            site_path = 'lib/site-packages'
        else:
            site_path = 'lib/python'+python_version_site+'/site-packages'

        # add platform specific python libs (.so/.pyd) to python module
        lib_path = os.path.join(esys_bin, 'esys')
        esys_bin_strip_len = len(esys_bin) + 1
        data_files = [
            (os.path.join(site_path, dp[esys_bin_strip_len:]), [os.path.join(dp, f)])
                for dp, dn, fns in os.walk(lib_path) for f in fns
        ]

        # add platform specific non-python libs to a separate site-packages folder
        lib_path = os.path.join(esys_bin, 'lib')
        self._updateBuildvars(os.path.join(lib_path, 'buildvars'))
        data_files.append(
            (
                os.path.join(site_path, esys_lib),
                [os.path.join(dp, f) for dp, dn, fns in os.walk(lib_path) for f in fns if f != 'buildvars']
            )
        )
        self.distribution.include(data_files=data_files)
        setuptools.command.install.install.run(self)

install_requires = ['matplotlib', 'netCDF4', 'scipy']
if python_version == '37':
    if plat.startswith('Windows'):
        install_requires.append('numpy==1.15.4')
    else:
        install_requires.append('numpy')
    install_requires.append('sympy==1.1.1')
else:
    install_requires.append('numpy')
    install_requires.append('sympy')

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name='esys-escript',
    version=esys_version,
    description='numerical modelling library using the finite element method',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://esys-escript.github.io',
    python_requires='>=3.7',
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    entry_points = {
        'console_scripts': [
            'run-escript=esys.escript.run_escript:main',
            'runmodel=esys.escriptcore.runmodel:main'
        ]
    },
    cmdclass = { 'install': esys_install },
    setup_requires = ['requests']
)




# -*- coding: utf-8 -*-

##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

"""
Test suite for nonlinearPDEs class

"""

__author__="Jaco du Plessis"

from esys.escript import NonlinearPDE, whereZero, grad, sin, cos, symmetric, matrixmult, FunctionOnBoundary, HAVE_SYMBOLS
if HAVE_SYMBOLS:
    from esys.escript import Symbol
import numpy
import esys.escriptcore.utestselect as unittest
from esys.escript.linearPDEs import IllegalCoefficient,IllegalCoefficientValue
import numpy as np
from esys.escript.pdetools import Locator


class Test_nonLinearPDEs(unittest.TestCase):
    DEBUG=False
    VERBOSE=False   
    
@unittest.skipIf(not HAVE_SYMBOLS, 'sympy not available')
class Test_nlpde(Test_nonLinearPDEs):
    def test_run(self):
        #test just to confirm nlpde works   
        u=Symbol('u', dim=self.domain.getDim())
        nlpde = NonlinearPDE(self.domain, u)
        x=self.domain.getX()
        gammaD=whereZero(x[0])+whereZero(x[1])
        nlpde.setValue(X=grad(u), Y=5*u, q=gammaD, r=1)
        v=nlpde.getSolution(u=1)

    def test_setVals1eq(self):
        #test setting Coefficients with 1 equation
        dim=self.domain.getDim()
        u=Symbol('u', dim=dim)
        nlpde = NonlinearPDE(self.domain, u)
        x=self.domain.getX()
        gammaD=whereZero(x[0])+whereZero(x[1])
        nlpde.setValue(X=grad(u), Y=5*u, q=gammaD, r=1)
        A=nlpde.getCoefficient("A")
        B=nlpde.getCoefficient("B")
        C=nlpde.getCoefficient("C")
        D=nlpde.getCoefficient("D")
        if dim==2:
            ATest=numpy.empty((2,2), dtype=object)
            ATest[0]=1,0
            ATest[1]=0,1
            BTest=numpy.empty((2,), dtype=object)
            BTest[0]=0
            BTest[1]=0
            CTest=BTest
            # Use dtype=int to match sympy Zero type from subtraction
            self.assertTrue(A-ATest==Symbol(numpy.zeros((2,2), dtype=int)))
            self.assertTrue(B-BTest==Symbol(numpy.zeros((2,), dtype=int)))
            self.assertTrue(C-CTest==Symbol(numpy.zeros((2,), dtype=int)))
            temp=Symbol('temp')
            self.assertTrue(D-temp.subs(temp,5)==temp.subs(temp,0))
        else:
            ATest=numpy.empty((3,3), dtype=object)
            ATest[0]=1,0,0
            ATest[1]=0,1,0
            ATest[2]=0,0,1
            BTest=numpy.empty((3,), dtype=object)
            BTest[0]=0
            BTest[1]=0
            BTest[2]=0
            CTest=BTest
            # Use dtype=int to match sympy Zero type from subtraction
            self.assertTrue(A-ATest==Symbol(numpy.zeros((3,3), dtype=int)))
            self.assertTrue(B-BTest==Symbol(numpy.zeros((3,), dtype=int)))
            self.assertTrue(C-CTest==Symbol(numpy.zeros((3,), dtype=int)))
            temp=Symbol('temp')
            self.assertTrue(D-temp.subs(temp,5)==temp.subs(temp,0))

    def test_setVals2eq(self):
        #test setting Coefficients with 2 coeficients
        dim=self.domain.getDim()
        u = Symbol('u',(2,), dim=dim)
        q = Symbol('q', (2,2))
        nlpde = NonlinearPDE(self.domain, u, debug=0)
        x = self.domain.getX()
        gammaD=whereZero(x[1])*[1,1]#+whereZero(x[0])*[1,0]+whereZero(x[0]-1)*[1,0]  
        yconstraint = FunctionOnBoundary(self.domain).getX()[1]
        nlpde.setValue(X=grad(u),q=gammaD,Y=[-50,-50]*u)
        A=nlpde.getCoefficient("A")
        B=nlpde.getCoefficient("B")
        C=nlpde.getCoefficient("C")
        D=nlpde.getCoefficient("D")
        if dim==2:
            ATest=numpy.empty((2,2,2,2),dtype=object)
            ATest[0]=(((1,0),(0,0)),((0,1),(0,0)))
            ATest[1]=(((0,0),(1,0)),((0,0),(0,1)))
            BTest=numpy.empty((2,2,2),dtype=object)
            BTest[0]=0
            BTest[1]=0
            CTest=BTest
            DTest=numpy.empty((2,2),dtype=object)
            DTest[0]=(-50,0)
            DTest[1]=(0,-50)
            self.assertTrue(numpy.ndarray.__eq__(ATest, A).all())
            self.assertTrue(numpy.ndarray.__eq__(BTest, B).all())
            self.assertTrue(numpy.ndarray.__eq__(CTest, C).all())
            self.assertTrue(numpy.ndarray.__eq__(DTest, D).all())
        else:
            ATest=numpy.empty((2,3,2,3),dtype=object)
            ATest[0]=(((1,0,0),(0,0,0)),((0,1,0),(0,0,0)),((0,0,1),(0,0,0)))
            ATest[1]=(((0,0,0),(1,0,0)),((0,0,0),(0,1,0)),((0,0,0),(0,0,1)))
            #ATest[1]=(((0,0,1),(1,0,0)),((0,0),(0,1)))
            BTest=numpy.empty((2,3,2),dtype=object)
            BTest[0]=0
            BTest[1]=0
            CTest=numpy.empty((2,2,3),dtype=object)
            CTest[0]=0
            CTest[1]=0
            DTest=numpy.empty((2,2),dtype=object)
            DTest[0]=(-50,0)
            DTest[1]=(0,-50)
            self.assertTrue(numpy.ndarray.__eq__(BTest, B).all())
            self.assertTrue(numpy.ndarray.__eq__(CTest, C).all())
            self.assertTrue(numpy.ndarray.__eq__(DTest, D).all())

    def test_DimAndShape1eq(self):
        dim=self.domain.getDim()
        if dim==3:
            u = Symbol('u', dim=2)
        else:
            u = Symbol('u', dim=3)
        nlpde = NonlinearPDE(self.domain, u, debug=0)
        args=dict(X=grad(u), Y=5*u)
        self.assertRaises(IllegalCoefficientValue, nlpde.setValue,**args)
        u = Symbol('u', dim=dim)
        args=dict(X=u, Y=5*u)
        self.assertRaises(IllegalCoefficientValue, nlpde.setValue,**args)
        args=dict(X=grad(u), Y=5*grad(u))
        self.assertRaises(IllegalCoefficientValue, nlpde.setValue,**args)
        #args=dict(q=u)
        #self.assertRaises(IllegalCoefficientValue, nlpde.setValue,**args)
    
    def test_DimAndShape2eq(self):
        dim=self.domain.getDim()
        u = Symbol('u',(2,), dim=dim)
        nlpde = NonlinearPDE(self.domain, u, debug=0)
        args=dict(X=grad(u), Y=5*u[0])
        self.assertRaises(IllegalCoefficientValue, nlpde.setValue,**args)
        args=dict(X=grad(u[0]), Y=5*u)
        self.assertRaises(IllegalCoefficientValue, nlpde.setValue,**args)

    def test_setUnknownPeram(self):
        dim=self.domain.getDim()
        u = Symbol('u',(2,), dim=dim)
        nlpde = NonlinearPDE(self.domain, u, debug=0)
        args=dict(k=0,f=8)  
        self.assertRaises(IllegalCoefficient,nlpde.setValue,**args)

    def test_yDirection(self):
        dim=self.domain.getDim()
        if dim==3:
            return
        u = Symbol('u',(2,), dim=dim)
        q = Symbol('q', (2,2))
        theta = Symbol('theta')
        theta=3.141/6
        q[0,0]=cos(theta)
        q[0,1]=-sin(theta)
        q[1,0]=sin(theta)
        q[1,1]=cos(theta)
        sigma = Symbol('sigma',(2,2))
        p = NonlinearPDE(self.domain, u, debug=0)
        epsilon = symmetric(grad(u))
     #   epsilon = matrixmult(matrixmult(q,epsilon0),q.transpose(1))
        c00=10;c01=8;c05=0
        c01=8;c11=10;c15=0
        c05=0;c15=0;c55=1
        sigma[0,0]=c00*epsilon[0,0]+c01*epsilon[1,1]+c05*2*epsilon[1,0]
        sigma[1,1]=c01*epsilon[0,0]+c11*epsilon[1,1]+c15*2*epsilon[1,0]
        sigma[0,1]=c05*epsilon[0,0]+c15*epsilon[1,1]+c55*2*epsilon[1,0]
        sigma[1,0]=sigma[0,1]
     #   sigma0=matrixmult(matrixmult(q.transpose(1),epsilon),q)
        x = self.domain.getX()
        gammaD=whereZero(x[1])*[1,1]#+whereZero(x[0])*[1,0]+whereZero(x[0]-1)*[1,0]  
        yconstraint = FunctionOnBoundary(self.domain).getX()[1]
        p.setValue(X=sigma,q=gammaD,y=[-50,0]*whereZero(yconstraint-1),r=[1,1])
        v = p.getSolution(u=[0,0])
        x=np.ndarray((2,))
        x[0]=0.5
        x[1]=0.5
        loc=Locator(v.getFunctionSpace(),x)
        valAtX=loc(v)
        self.assertTrue(valAtX[0]>10*valAtX[1])





##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# Development 2012-2013 by School of Earth Sciences
# Development from 2014 by Centre for Geoscience Computing (GeoComp)
# Development from 2019 by School of Earth and Environmental Sciences
#
##############################################################################

"""Classes and tools that form the basis of the escript system.
Specific solvers and domains are found in their respective packages."""


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

import os, sys

esys_packages = __name__.split('.')
esys_paths = os.path.dirname(os.path.abspath(__file__)).split(os.sep)
__buildinfo_file__ = None
package_prefix = os.sep.join(esys_paths[:-len(esys_packages)])
esys_lib_path = os.sep.join([package_prefix, 'esys_escript_lib'])
# The pip/wheel layout ships escript's shared libs and a buildvars.in
# template in a sibling esys_escript_lib/ directory and carries no RPATH, so
# it must bootstrap the dynamic library path here. conda and system installs
# resolve their libs via RPATH (or a system path) and have no
# esys_escript_lib/ directory, so skip this bootstrap for them -- otherwise
# bare `import esys` fails trying to read esys_escript_lib/buildvars.in.
if esys_paths[-(len(esys_packages)+1)] == 'site-packages' and os.path.isdir(esys_lib_path):
    # check the library path for a pip install
    if os.name == 'nt':
        lib_env_name = 'PATH'
        cmd, env1, env2 = 'set', '%', '%'
    else:
        lib_env_name = 'LD_LIBRARY_PATH'
        cmd, env1, env2 = 'export', '$', ''
    buildinfo_name = 'buildvars'
    __buildinfo_file__ = os.path.join(esys_lib_path, buildinfo_name)
    # check for the buildvars file
    if not os.path.exists(__buildinfo_file__):
        print('\ncreating {bv} file using {bv}.in template\n'.format(bv=buildinfo_name))
        with open(__buildinfo_file__+'.in', 'r') as f:
            lines = f.readlines()
        txt = ''.join(lines) % { 'python': sys.executable, 'prefix': package_prefix,
            'inc_prefix': '', 'lib_prefix': esys_lib_path }
        with open(__buildinfo_file__, 'w') as f:
            f.write(txt)
    lib_env_paths = os.environ.get(lib_env_name, '').split(os.pathsep)
    # re-start python if needed
    if esys_lib_path in lib_env_paths:
        print('\nusing escript library path: {}\n'.format(esys_lib_path))
    else:
        # set library path and restart python if not set
        new_lib_env_paths = os.sep.join([esys_lib_path] + lib_env_paths)
        print('\nwarning: escript library path not set')
        print('{} {}={}{}{}{}{}'.format(cmd, lib_env_name, esys_lib_path, os.pathsep, env1, lib_env_name, env2))
        env = os.environ
        env[lib_env_name] = os.pathsep.join(lib_env_paths+[esys_lib_path])
        if len(''.join(sys.argv)) > 0:
            print('re-starting python with new library path...\n')
            os.execve(sys.executable, [sys.executable]+sys.argv, env)
            sys.exit(0)
        else:
            # no args, don't restart an interactive session
            raise Exception('please exit python shell, set escript library path, then re-start')

from esys.escriptcore.escriptcpp import *
from esys.escriptcore.start import HAVE_SYMBOLS
from esys.escriptcore.util import *
from esys.escriptcore.interpolation import InterpolationTable
from esys.escriptcore.nonlinearPDE import NonlinearPDE
from esys.escriptcore.datamanager import DataManager
if HAVE_SYMBOLS:
    from esys.escriptcore.symboliccore import *
from . import minimizer

# Import MPI domain coupler only if mpi4py is available
try:
    from mpi4py import MPI as _MPI
    from .domaincoupler import MPIDomainArray, DataCoupler
    HAVE_MPI4PY = True
except ImportError:
    HAVE_MPI4PY = False
    # Provide placeholder classes for documentation purposes
    MPIDomainArray = None
    DataCoupler = None

import logging
logging.basicConfig(format='%(name)s: %(message)s', level=logging.INFO)

__all__=[x for x in dir() if not x.startswith('internal_') and not x.startswith('Internal_') and not x.startswith('__') and not str(type(eval(x))).find('module')>=0]





##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

# This is a template configuration file for escript on Debian/Ubuntu Linux.
# Refer to installation.md for usage instructions.
#
# Prerequisites:
#   sudo apt-get install python3-dev python3-numpy python3-scipy python3-sympy
#   sudo apt-get install scons cmake g++ gfortran
#   sudo apt-get install libboost-python-dev libboost-numpy-dev libboost-random-dev libboost-iostreams-dev
#   sudo apt-get install libnetcdf-dev libsilo-dev libhdf5-serial-dev
#   sudo apt-get install libsuitesparse-dev liblapacke-dev libmumps-seq-dev
#   sudo apt-get install libmetis-dev libparmetis-dev zlib1g-dev
#   sudo apt-get install libopenmpi-dev openmpi-bin python3-mpi4py
#
# For no-MPI build, use debian_nompi_options.py instead.

import subprocess

escript_opts_version = 203

# Python configuration - auto-detect version
pythoncmd = 'python3'
p = subprocess.run([pythoncmd, '-V'], capture_output=True, text=True)
subversion = p.stdout.split(' ')[1].split('.')[1]
print(f"Python subversion = {subversion}")

pythonlibname = f'python3.{subversion}'
pythonlibpath = '/usr/lib/x86_64-linux-gnu/'
pythonincpath = f'/usr/include/python3.{subversion}'

# Compiler settings
openmp = True
werror = False
cxx_extra = ['-O3', '-fdiagnostics-color=always', '-fstack-protector-strong', '-Wformat', '-Werror=format-security']

# Boost configuration
boost_prefix = ['/usr/include', '/usr/lib/x86_64-linux-gnu/']
boost_libs = [f'boost_python3{subversion}', f'boost_numpy3{subversion}', 'boost_random']

# MPI configuration
# Note: OpenMPI 5.x removed mpi_cxx and open-rte libraries; just 'mpi' is needed
mpi = 'OPENMPI'
mpi_prefix = ['/usr/include/x86_64-linux-gnu/openmpi', '/usr/lib/x86_64-linux-gnu/openmpi']
mpi_libs = ['mpi']
mpi4py = True

# Solver configuration
paso = True
build_trilinos = 'check'

# UMFPACK direct solver
umfpack = True
umfpack_prefix = ['/usr/include/suitesparse', '/usr/lib']
umfpack_libs = ['umfpack', 'blas', 'amd']

# MUMPS sequential solver
mumps_seq = True
mumps_seq_prefix = ['/usr/include', '/usr/lib/x86_64-linux-gnu']
mumps_seq_libs = ['dmumps_seq', 'zmumps_seq', 'mumps_common_seq', 'mpiseq_seq', 'pord_seq']

# METIS graph partitioning (for Trilinos)
metis = True
metis_prefix = ['/usr/include', '/usr/lib/x86_64-linux-gnu']
metis_libs = ['metis']

# ParMETIS parallel graph partitioning (requires MPI)
parmetis = True
# Note: On Ubuntu 24.04+, parmetis headers and libs are in /usr/include and /usr/lib
# (no longer in /usr/include/parmetis or /usr/lib/x86_64-linux-gnu)
parmetis_prefix = ['/usr/include', '/usr/lib']
parmetis_libs = ['parmetis']

# LAPACK configuration
lapack = 'auto'
lapack_prefix = ['/usr/include', '/usr/lib/x86_64-linux-gnu']
lapack_libs = ['lapacke']

# File I/O libraries
netcdf = True
netcdf_prefix = ['/usr/include', '/usr/lib/x86_64-linux-gnu']
netcdf_libs = ['netcdf_c++4', 'netcdf']

silo = True
silo_prefix = ['/usr/include', '/usr/lib/x86_64-linux-gnu/hdf5/serial', '/usr/lib/x86_64-linux-gnu']
silo_libs = ['siloh5', 'hdf5']

hdf5 = True
hdf5_prefix = ['/usr/include/hdf5/serial', '/usr/lib/x86_64-linux-gnu/hdf5/serial', '/usr/lib/x86_64-linux-gnu']
hdf5_libs = ['hdf5_serial_cpp']

# Compression support
zlib = True
zlib_prefix = ['/usr/include', '/usr/lib/x86_64-linux-gnu']
zlib_libs = ['z']

# Optional features
sympy = True

# Clean up temporary variables (SCons treats all module-level variables as options)
del p, subversion



escript_opts_version=203


boost_libs=['boost_python39','boost_numpy39']
boost_prefix=['/usr/include/','/usr/lib/x86_64-linux-gnu/']
build_trilinos='make'
lapack=1
lapack_prefix=['/usr/include/x86_64-linux-gnu','/usr/include/x86_64-linux-gnu/']
mpi4py=1
mpi='OPENMPI'
mpi_prefix='/usr/lib/x86_64-linux-gnu/openmpi/'
mumps=0
mumps_prefix=['/usr/include/x86_64-linux-gnu','/usr/include/x86_64-linux-gnu/']
openmp=1
paso=1
pythoncmd='python3'
pythonlibpath='/usr/lib/x86_64-linux-gnu/'
pythonincpath='/usr/include/python3.9/'
pythonlibname='python3.9'
silo=1
silo_libs=['siloh5','hdf5_serial']
silo_prefix=['/usr/include','/usr/lib/x86_64-linux-gnu/']
umfpack=1
umfpack_prefix=['/usr/include/suitesparse','/usr/lib/x86_64-linux-gnu/']
werror=0



#!/usr/bin/python3
from esys.escript import *
from esys.finley import ReadGmsh
import unittest

class Test_Finley_GMSHReader(unittest.TestCase):
    ELECTRODES = [(-64., 0., 0.), (-32., 0., 0.), ( 0.0, 0., 0.), ( 32., 0., 0.), ( 64., 0., 0.)]
    DIRACTAGS = ["e"+str(st) for st in range(5)]
    VOL=2000000.
    AREA=100000.
    DIRACCOUNT=5.0
    def test_Version22_addDirac(self):
        """
        mesh version 2.2 and addition of Dirac point when reading
        it is known that version 2.2 does not contain "Physical Point"
        
        """
        MSHFILE="meshDirac_22.msh"
        domain=ReadGmsh(MSHFILE, 3, optimize=True, diracPoints = self.ELECTRODES, diracTags=self.DIRACTAGS)
        self.checkDomain(domain)
        self.checkTags(domain)
        

    def test_Version41(self):
        MSHFILE="meshDirac_41.msh"
        domain=ReadGmsh(MSHFILE, 3, optimize=True)
        self.checkDomain(domain)
        self.checkTags(domain)

    def test_Version41_addDirac(self):
        MSHFILE="meshDirac_41.msh"
        domain=ReadGmsh(MSHFILE, 3, optimize=True,  diracPoints = self.ELECTRODES, diracTags=self.DIRACTAGS)
        self.checkDomain(domain)
        self.checkTags(domain)


    def test_Version42(self):
        MSHFILE="meshDirac_42.msh"
        domain=ReadGmsh(MSHFILE, 3, optimize=True)
        self.checkDomain(domain)
        self.checkTags(domain)

    def test_Version42_addDirac(self):
        MSHFILE="meshDirac_42.msh"
        domain=ReadGmsh(MSHFILE, 3, optimize=True,  diracPoints = self.ELECTRODES, diracTags=self.DIRACTAGS)
        self.checkDomain(domain)
        self.checkTags(domain)


    def test_Version44(self):

        MSHFILE="meshDirac_44.msh"
        domain=ReadGmsh(MSHFILE, 3, optimize=True)
        self.checkDomain(domain)
        self.checkTags(domain)


    def test_Version44_addDirac(self):

        MSHFILE="meshDirac_44.msh"
        domain=ReadGmsh(MSHFILE, 3, optimize=True,  diracPoints = self.ELECTRODES, diracTags=self.DIRACTAGS)
        self.checkDomain(domain)
        self.checkTags(domain)

    def checkDomain(self, domain):
        one=Scalar(1., Function(domain))
        V=integrate(one)
        self.assertAlmostEqual(V,  self.VOL, msg="wrong volume")
        
        one=Scalar(1., FunctionOnBoundary(domain))
        A=integrate(one)
        self.assertAlmostEqual(A,  self.AREA, msg="wrong surface area")

        one=Scalar(1., DiracDeltaFunctions(domain))
        D=integrate(one)
        self.assertAlmostEqual(D,  self.DIRACCOUNT, msg="wrong dirac point count")

    def checkTags(self, domain, alternative_dirac_tags=None):
        
        mesh_tags=getTagNames(domain)
        self.assertListEqual(mesh_tags, ['back', 'bottom', 'e0', 'e1', 'e2', 'e3', 'e4', 'front', 'left', 'right', 'top', 'volume'])       
        
        tagsF=Function(domain).getListOfTags() 
        tagsF.sort()
        self.assertListEqual(tagsF, [1], 'wrong volume tags')

        tagsFBC=FunctionOnBoundary(domain).getListOfTags()
        tagsFBC.sort()
        self.assertListEqual(tagsFBC, [2, 3, 4, 5, 6, 7], 'wrong surface tags')

        tagsD=DiracDeltaFunctions(domain).getListOfTags() 
        tagsD.sort()
        # this tag ids could be different when tags are added while reading
        if alternative_dirac_tags:
            self.assertListEqual(tagsD, alternative_dirac_tags, 'wrong Dirac point tags')
        else:    
            self.assertListEqual(tagsD, [8, 9, 10, 11, 12], 'wrong Dirac point tags')
        
        

if __name__ == '__main__':
    unittest.main()




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

"""
Comprehensive test suite for MPIDomainArray and DataCoupler classes.

This module provides extensive testing for MPI-based domain decomposition and
inter-domain communication in escript. The test suite validates both the
communicator topology infrastructure (MPIDomainArray) and the data coupling
operations (DataCoupler) across multiple domain types and function spaces.

Test Coverage
-------------
The test suite includes:

**MPIDomainArray Tests:**
    - Basic 2-domain topology creation
    - Multiple subdomains per domain
    - Communicator hierarchy validation

**DataCoupler Point-to-Point Communication:**
    - send() and receive() operations for scalar and vector Data
    - Bidirectional exchange() operations
    - Communication across all supported function spaces

**DataCoupler Collective Operations on Data Objects:**
    - broadcast() operations from root domain to all domains
    - allreduce() operations with SUM operator

**DataCoupler Collective Operations on Scalar Values:**
    - broadcast_value() for Python scalar distribution
    - allreduce_value() with SUM and MAX operators

**Domain Types Tested:**
    - esys.ripley (Rectangle and Brick)
    - esys.finley (Rectangle and Brick)
    - esys.speckley (Rectangle and Brick with order=2)
    - Note: esys.oxley excluded due to deadlock issues (see GitHub issue #123)

**Function Spaces Tested:**
    - Solution
    - ContinuousFunction
    - ReducedSolution
    - Function
    - ReducedFunction
    - FunctionOnBoundary
    - ReducedFunctionOnBoundary
    - DiracDeltaFunctions

**Data Types Tested:**
    - Scalar Data objects
    - Vector Data objects (2D and 3D)
    - Python scalar values

Requirements
------------
These tests require MPI with at least 2 processes. Some tests require exactly
4 processes for multiple subdomain validation.

Usage
-----
Run the complete test suite with::

    run-escript -p 4 run_domaincoupler.py

Or run individual test classes with::

    run-escript -p 4 -m unittest run_domaincoupler.Test_MPIDomainArray

Notes
-----
- Speckley domains do not support ReducedSolution, FunctionOnBoundary, or
  ReducedFunctionOnBoundary. These combinations are automatically skipped.
- All tests use MPI collective operations, requiring all ranks to participate.
- Tests validate results on all participating ranks using collective reduction
  operations (sup, inf).
- Dirac points are automatically added to all domains to enable testing of
  DiracDeltaFunctions function space.

See Also
--------
MPIDomainArray : MPI communicator topology for domain decomposition
DataCoupler : Inter-domain communication interface
"""

__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""

from esys.escript import *
from esys.escript import MPIDomainArray, DataCoupler
from esys.escript.linearPDEs import LinearPDE
import esys.escriptcore.utestselect as unittest
from esys.escriptcore.testing import *
import numpy as np

try:
    from mpi4py import MPI
    HAVE_MPI = True
    WORLD_SIZE = MPI.COMM_WORLD.Get_size()
    WORLD_RANK = MPI.COMM_WORLD.Get_rank()
except ImportError:
    HAVE_MPI = False
    WORLD_SIZE = 1
    WORLD_RANK = 0

# Skip all tests if MPI is not available or the rank count is not compatible
# with the hardcoded numDomains=2 used throughout this file.  MPIDomainArray
# requires WORLD_SIZE to be a multiple of numDomains, so e.g. 3 ranks fails
# with "Requested ranks 2 does not match available ranks 3".  Even rank
# counts >= 2 (2, 4, 6, ...) work.
SKIP_MPI_TESTS = not HAVE_MPI or WORLD_SIZE < 2 or WORLD_SIZE % 2 != 0


# Test configurations: (domain_module, domain_class, dimensions, resolution)
DOMAIN_CONFIGS = [
    ('esys.ripley', 'Rectangle', 2, {'n0': 10, 'n1': 10}),
    ('esys.ripley', 'Brick', 3, {'n0': 8, 'n1': 8, 'n2': 8}),
    ('esys.finley', 'Rectangle', 2, {'n0': 10, 'n1': 10}),
    ('esys.finley', 'Brick', 3, {'n0': 6, 'n1': 6, 'n2': 6}),
    ('esys.speckley', 'Rectangle', 2, {'order': 2, 'n0': 10, 'n1': 10}),
    ('esys.speckley', 'Brick', 3, {'order': 2, 'n0': 6, 'n1': 6, 'n2': 6}),
]

# Oxley commented out - may cause deadlocks with DataCoupler operations
# try:
#     import esys.oxley
#     DOMAIN_CONFIGS.extend([
#         ('esys.oxley', 'Rectangle', 2, {'n0': 10, 'n1': 10}),
#         ('esys.oxley', 'Brick', 3, {'n0': 6, 'n1': 6, 'n2': 6}),
#     ])
# except ImportError:
#     pass


def create_domain(domain_module_name, domain_class_name, params, comm):
    """Helper to create a domain instance with Dirac points."""
    import importlib
    module = importlib.import_module(domain_module_name)
    domain_class = getattr(module, domain_class_name)

    # Add Dirac points for testing DiracDeltaFunctions
    dirac_params = params.copy()
    if 'n2' in params:
        # 3D domain - add two Dirac points
        dirac_params['diracPoints'] = [(0.5, 0.5, 0.5), (0.8, 0.8, 0.8)]
        dirac_params['diracTags'] = ['point1', 'point2']
    elif 'n1' in params:
        # 2D domain - add two Dirac points
        dirac_params['diracPoints'] = [(0.5, 0.5), (0.8, 0.8)]
        dirac_params['diracTags'] = ['point1', 'point2']

    return domain_class(**dirac_params, comm=comm)


@unittest.skipIf(SKIP_MPI_TESTS, "MPI with an even number of processes (>= 2) required")
class Test_MPIDomainArray(unittest.TestCase):
    """Tests for MPIDomainArray communicator topology."""

    def test_basic_topology_2domains(self):
        """Test basic 2-domain topology creation."""
        num_domains = 2
        domain_array = MPIDomainArray(numDomains=num_domains, comm=MPI.COMM_WORLD)

        # Check basic properties (all ranks)
        self.assertEqual(domain_array.numDomains, num_domains)
        self.assertIsNotNone(domain_array.getDomainComm())
        self.assertIsNotNone(domain_array.getSubdomainComm())

        # Check domain index is valid
        domain_idx = domain_array.getDomainIndex()
        self.assertIn(domain_idx, range(num_domains))

        # Check subdomain communicator connects all domains
        subdomain_comm = domain_array.getSubdomainComm()
        self.assertEqual(subdomain_comm.Get_size(), num_domains)

    @unittest.skipIf(WORLD_SIZE < 4, "Need at least 4 processes")
    def test_topology_with_multiple_subdomains(self):
        """Test topology with multiple subdomains per domain."""
        num_domains = 2
        num_subdomains = WORLD_SIZE // 2
        domain_array = MPIDomainArray(numDomains=num_domains,
                                      numSubDomains=num_subdomains,
                                      comm=MPI.COMM_WORLD)

        # Check domain communicator has correct size
        domain_comm = domain_array.getDomainComm()
        self.assertEqual(domain_comm.Get_size(), num_subdomains)

        # Check subdomain communicator connects domains
        subdomain_comm = domain_array.getSubdomainComm()
        self.assertEqual(subdomain_comm.Get_size(), num_domains)


@unittest.skipIf(SKIP_MPI_TESTS, "MPI with an even number of processes (>= 2) required")
class Test_DataCoupler_PointToPoint(unittest.TestCase):
    """Tests for DataCoupler point-to-point communication across all domains and function spaces."""

    def test_send_receive_scalar(self):
        """Test sending and receiving scalar Data across all domains and function spaces."""
        num_domains = 2
        domain_array = MPIDomainArray(numDomains=num_domains, comm=MPI.COMM_WORLD)
        my_domain_idx = domain_array.getDomainIndex()
        coupler = DataCoupler(domain_array)

        # Test each domain configuration
        for domain_module, domain_class, dim, params in DOMAIN_CONFIGS:
            with self.subTest(domain=f"{domain_module}.{domain_class}"):
                domain = create_domain(domain_module, domain_class, params,
                                      domain_array.getDomainComm())

                # Test each function space
                for fs_name, fs_constructor in [
                    ('Solution', Solution),
                    ('ContinuousFunction', ContinuousFunction),
                    ('ReducedSolution', ReducedSolution),
                    ('Function', Function),
                    ('ReducedFunction', ReducedFunction),
                    ('FunctionOnBoundary', FunctionOnBoundary),
                    ('ReducedFunctionOnBoundary', ReducedFunctionOnBoundary),
                    ('DiracDeltaFunctions', DiracDeltaFunctions),
                ]:
                    # Skip unsupported speckley function space combinations
                    if domain_module == 'esys.speckley' and fs_name in [
                        'ReducedSolution', 'FunctionOnBoundary', 'ReducedFunctionOnBoundary'
                    ]:
                        continue

                    with self.subTest(function_space=fs_name):
                        fs = fs_constructor(domain)

                        # All ranks participate
                        if my_domain_idx == 0:
                            # Domain 0 sends
                            data = Scalar(42.0, fs)
                            coupler.send(data, dest_domain_index=1, tag=100)
                        elif my_domain_idx == 1:
                            # Domain 1 receives
                            received = coupler.receive(fs, source_domain_index=0, tag=100)

                            # All ranks in domain 1 verify (sup is collective)
                            received_val = sup(received)
                            self.assertAlmostEqual(received_val, 42.0, places=10,
                                                 msg=f"{domain_module}.{domain_class} {fs_name}")

    def test_send_receive_vector(self):
        """Test sending and receiving vector Data across all domains and function spaces."""
        num_domains = 2
        domain_array = MPIDomainArray(numDomains=num_domains, comm=MPI.COMM_WORLD)
        my_domain_idx = domain_array.getDomainIndex()
        coupler = DataCoupler(domain_array)

        # Test each domain configuration
        for domain_module, domain_class, dim, params in DOMAIN_CONFIGS:
            with self.subTest(domain=f"{domain_module}.{domain_class}"):
                domain = create_domain(domain_module, domain_class, params,
                                      domain_array.getDomainComm())

                # Test each function space
                for fs_name, fs_constructor in [
                    ('Solution', Solution),
                    ('ContinuousFunction', ContinuousFunction),
                    ('ReducedSolution', ReducedSolution),
                    ('Function', Function),
                    ('ReducedFunction', ReducedFunction),
                    ('FunctionOnBoundary', FunctionOnBoundary),
                    ('ReducedFunctionOnBoundary', ReducedFunctionOnBoundary),
                    ('DiracDeltaFunctions', DiracDeltaFunctions),
                ]:
                    # Skip unsupported speckley function space combinations
                    if domain_module == 'esys.speckley' and fs_name in [
                        'ReducedSolution', 'FunctionOnBoundary', 'ReducedFunctionOnBoundary'
                    ]:
                        continue

                    with self.subTest(function_space=fs_name):
                        fs = fs_constructor(domain)

                        # All ranks participate
                        if my_domain_idx == 0:
                            # Domain 0 sends vector
                            if dim == 2:
                                data = Vector([1.0, 2.0], fs)
                            else:  # dim == 3
                                data = Vector([1.0, 2.0, 3.0], fs)
                            coupler.send(data, dest_domain_index=1, tag=200)
                        elif my_domain_idx == 1:
                            # Domain 1 receives
                            received = coupler.receive(fs, source_domain_index=0, tag=200)

                            # All ranks verify (sup is collective)
                            if dim == 2:
                                self.assertAlmostEqual(sup(received[0]), 1.0, places=10,
                                                     msg=f"{domain_module}.{domain_class} {fs_name} [0]")
                                self.assertAlmostEqual(sup(received[1]), 2.0, places=10,
                                                     msg=f"{domain_module}.{domain_class} {fs_name} [1]")
                            else:  # dim == 3
                                self.assertAlmostEqual(sup(received[0]), 1.0, places=10,
                                                     msg=f"{domain_module}.{domain_class} {fs_name} [0]")
                                self.assertAlmostEqual(sup(received[1]), 2.0, places=10,
                                                     msg=f"{domain_module}.{domain_class} {fs_name} [1]")
                                self.assertAlmostEqual(sup(received[2]), 3.0, places=10,
                                                     msg=f"{domain_module}.{domain_class} {fs_name} [2]")

    def test_exchange_bidirectional(self):
        """Test bidirectional exchange across all domains and function spaces."""
        num_domains = 2
        domain_array = MPIDomainArray(numDomains=num_domains, comm=MPI.COMM_WORLD)
        my_domain_idx = domain_array.getDomainIndex()
        coupler = DataCoupler(domain_array)

        # Test each domain configuration
        for domain_module, domain_class, dim, params in DOMAIN_CONFIGS:
            with self.subTest(domain=f"{domain_module}.{domain_class}"):
                domain = create_domain(domain_module, domain_class, params,
                                      domain_array.getDomainComm())

                # Test each function space
                for fs_name, fs_constructor in [
                    ('Solution', Solution),
                    ('ContinuousFunction', ContinuousFunction),
                    ('ReducedSolution', ReducedSolution),
                    ('Function', Function),
                    ('ReducedFunction', ReducedFunction),
                    ('FunctionOnBoundary', FunctionOnBoundary),
                    ('ReducedFunctionOnBoundary', ReducedFunctionOnBoundary),
                    ('DiracDeltaFunctions', DiracDeltaFunctions),
                ]:
                    # Skip unsupported speckley function space combinations
                    if domain_module == 'esys.speckley' and fs_name in [
                        'ReducedSolution', 'FunctionOnBoundary', 'ReducedFunctionOnBoundary'
                    ]:
                        continue

                    with self.subTest(function_space=fs_name):
                        fs = fs_constructor(domain)

                        # All ranks participate
                        if my_domain_idx == 0:
                            send_data = Scalar(100.0, fs)
                            received = coupler.exchange(send_data, fs, peer_domain_index=1, tag=300)
                            # Verify received data from domain 1
                            received_val = sup(received)
                            self.assertAlmostEqual(received_val, 200.0, places=10,
                                                 msg=f"{domain_module}.{domain_class} {fs_name}")
                        elif my_domain_idx == 1:
                            send_data = Scalar(200.0, fs)
                            received = coupler.exchange(send_data, fs, peer_domain_index=0, tag=300)
                            # Verify received data from domain 0
                            received_val = sup(received)
                            self.assertAlmostEqual(received_val, 100.0, places=10,
                                                 msg=f"{domain_module}.{domain_class} {fs_name}")


@unittest.skipIf(SKIP_MPI_TESTS, "MPI with an even number of processes (>= 2) required")
class Test_DataCoupler_Collective_Data(unittest.TestCase):
    """Tests for DataCoupler collective operations on Data objects."""

    def test_broadcast_data_scalar(self):
        """Test broadcasting scalar Data across all domains and function spaces."""
        num_domains = 2
        domain_array = MPIDomainArray(numDomains=num_domains, comm=MPI.COMM_WORLD)
        my_domain_idx = domain_array.getDomainIndex()
        coupler = DataCoupler(domain_array)

        # Test each domain configuration
        for domain_module, domain_class, dim, params in DOMAIN_CONFIGS:
            with self.subTest(domain=f"{domain_module}.{domain_class}"):
                domain = create_domain(domain_module, domain_class, params,
                                      domain_array.getDomainComm())

                # Test each function space
                for fs_name, fs_constructor in [
                    ('Solution', Solution),
                    ('ContinuousFunction', ContinuousFunction),
                    ('ReducedSolution', ReducedSolution),
                    ('Function', Function),
                    ('ReducedFunction', ReducedFunction),
                    ('FunctionOnBoundary', FunctionOnBoundary),
                    ('ReducedFunctionOnBoundary', ReducedFunctionOnBoundary),
                    ('DiracDeltaFunctions', DiracDeltaFunctions),
                ]:
                    # Skip unsupported speckley function space combinations
                    if domain_module == 'esys.speckley' and fs_name in [
                        'ReducedSolution', 'FunctionOnBoundary', 'ReducedFunctionOnBoundary'
                    ]:
                        continue

                    with self.subTest(function_space=fs_name):
                        fs = fs_constructor(domain)

                        # All ranks participate in broadcast
                        if my_domain_idx == 0:
                            # Domain 0 broadcasts
                            data = Scalar(99.0, fs)
                            result = coupler.broadcast(data, root_domain_index=0)
                        else:
                            # Other domains receive
                            result = coupler.broadcast(function_space=fs, root_domain_index=0)

                        # All ranks verify
                        result_val = sup(result)
                        self.assertAlmostEqual(result_val, 99.0, places=10,
                                             msg=f"{domain_module}.{domain_class} {fs_name}")

    def test_allreduce_data_sum(self):
        """Test all-reduce SUM on Data across all domains and function spaces."""
        num_domains = 2
        domain_array = MPIDomainArray(numDomains=num_domains, comm=MPI.COMM_WORLD)
        my_domain_idx = domain_array.getDomainIndex()
        coupler = DataCoupler(domain_array)

        # Test each domain configuration
        for domain_module, domain_class, dim, params in DOMAIN_CONFIGS:
            with self.subTest(domain=f"{domain_module}.{domain_class}"):
                domain = create_domain(domain_module, domain_class, params,
                                      domain_array.getDomainComm())

                # Test each function space
                for fs_name, fs_constructor in [
                    ('Solution', Solution),
                    ('ContinuousFunction', ContinuousFunction),
                    ('ReducedSolution', ReducedSolution),
                    ('Function', Function),
                    ('ReducedFunction', ReducedFunction),
                    ('FunctionOnBoundary', FunctionOnBoundary),
                    ('ReducedFunctionOnBoundary', ReducedFunctionOnBoundary),
                    ('DiracDeltaFunctions', DiracDeltaFunctions),
                ]:
                    # Skip unsupported speckley function space combinations
                    if domain_module == 'esys.speckley' and fs_name in [
                        'ReducedSolution', 'FunctionOnBoundary', 'ReducedFunctionOnBoundary'
                    ]:
                        continue

                    with self.subTest(function_space=fs_name):
                        print(domain_module, fs_name)
                        fs = fs_constructor(domain)

                        # All ranks participate - each domain contributes its index
                        data = Scalar(float(my_domain_idx), fs)
                        result = coupler.allreduce(data, op=MPI.SUM)

                        # All ranks verify
                        expected = sum(range(num_domains))  # 0 + 1 = 1
                        result_val = sup(result)
                        self.assertAlmostEqual(result_val, expected, places=10,
                                             msg=f"{domain_module}.{domain_class} {fs_name}")
                        print("DONE")

@unittest.skipIf(SKIP_MPI_TESTS, "MPI with an even number of processes (>= 2) required")
class Test_DataCoupler_Collective_Values(unittest.TestCase):
    """Tests for DataCoupler collective operations on scalar values (not Data objects)."""

    def test_broadcast_value_scalar(self):
        """Test broadcasting a scalar value."""
        num_domains = 2
        domain_array = MPIDomainArray(numDomains=num_domains, comm=MPI.COMM_WORLD)
        my_domain_idx = domain_array.getDomainIndex()
        coupler = DataCoupler(domain_array)

        # Create a dummy domain for isRootRank check
        domain = create_domain('esys.ripley', 'Rectangle', {'n0': 10, 'n1': 10},
                               domain_array.getDomainComm())

        # All ranks participate in broadcast
        if my_domain_idx == 0 and domain.isRootRank():
            result = coupler.broadcast_value(3.14159, root_domain_index=0)
        else:
            result = coupler.broadcast_value(root_domain_index=0)

        # All ranks verify
        self.assertAlmostEqual(result, 3.14159, places=5)

    def test_allreduce_value_sum(self):
        """Test all-reduce with SUM operation on scalars."""
        num_domains = 2
        domain_array = MPIDomainArray(numDomains=num_domains, comm=MPI.COMM_WORLD)
        my_domain_idx = domain_array.getDomainIndex()
        coupler = DataCoupler(domain_array)

        # All ranks participate
        my_value = float(my_domain_idx)
        result = coupler.allreduce_value(my_value, op=MPI.SUM)

        # All ranks verify
        expected = sum(range(num_domains))
        self.assertAlmostEqual(result, expected)

    def test_allreduce_value_max(self):
        """Test all-reduce with MAX operation."""
        num_domains = 2
        domain_array = MPIDomainArray(numDomains=num_domains, comm=MPI.COMM_WORLD)
        my_domain_idx = domain_array.getDomainIndex()
        coupler = DataCoupler(domain_array)

        # All ranks participate
        my_value = float(my_domain_idx)
        result = coupler.allreduce_value(my_value, op=MPI.MAX)

        # All ranks verify
        expected = float(num_domains - 1)
        self.assertAlmostEqual(result, expected)


# Create test suite
def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.makeSuite(Test_MPIDomainArray))
    test_suite.addTest(unittest.makeSuite(Test_DataCoupler_PointToPoint))
    test_suite.addTest(unittest.makeSuite(Test_DataCoupler_Collective_Data))
    test_suite.addTest(unittest.makeSuite(Test_DataCoupler_Collective_Values))
    return test_suite


if __name__ == '__main__':
    run_tests(__name__, exit_on_failure=True)




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

import logging
logging.basicConfig(format='%(name)s: %(message)s', level=logging.INFO)

from esys.escriptcore.pdetools import *



##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

__copyright__ = """Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__ = """Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__ = "https://github.com/LutzGross/esys-escript.github.io"

from esys.escript import *
from esys.escript.pdetools import MaskFromBoundaryTag
from esys.escript.linearPDEs import LinearPDE

try:
    from esys.finley import ReadGmsh
    HAVE_FINLEY = True
except ImportError:
    HAVE_FINLEY = False
from esys.weipa import saveVTK

if not HAVE_FINLEY:
    print("Finley module not available")
else:
    mkDir("output")
    # ... generate domain ...
    print("read in mesh")
    domain = ReadGmsh("inclusion.msh", 2, optimize=True)
    # ... set some parameters ...
    kappa = 1.
    q_c = 0.1
    q_out = 0.05
    T_ref = 15.

    # set source term:
    q_H=Scalar(0., Function(domain))
    q_H.setTaggedValue("Inclusion", q_c)
    # set flux term:
    q_S=Scalar(0., FunctionOnBoundary(domain))
    q_S.setTaggedValue("Left", q_out)
    q_S.setTaggedValue("Right", q_out)
    # ... set temperature at the bottom of the domain:
    # y=domain.getX()[1]
    # fixedT1 = whereZero(y - inf(y))
    fixedT = MaskFromBoundaryTag(domain, *('Bottom', ))
    # ... open PDE and set coefficients ...
    mypde = LinearPDE(domain)
    mypde.setSymmetryOn()
    mypde.setValue(A=kappa * kronecker(domain), Y=q_H, y=q_S, \
                   q=fixedT, r=T_ref)
    # ... calculate error of the PDE solution ...
    T = mypde.getSolution()
    print("Temperature is ", T)
    saveVTK("output/T.vtu", temperature=T)




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
##############################################################################

"""
Test suite for MPI communicator retrieval on Finley domains
"""

__copyright__ = """Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__ = """Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""

import esys.escriptcore.utestselect as unittest
from esys.escriptcore.testing import *
from test_mpiComm import Test_MPI_Comm_Retrieval

from esys.escript import getMPISizeWorld
from esys.finley import Rectangle, Brick, ReadMesh, ReadGmsh, LoadMesh
import os

# Test data paths
try:
    FINLEY_TEST_DATA = os.environ['FINLEY_TEST_DATA']
except KeyError:
    # Default to finley/test/python directory
    FINLEY_TEST_DATA = os.path.dirname(__file__)

FINLEY_TEST_MESH_PATH = os.path.join(FINLEY_TEST_DATA, "data_meshes")

try:
    FINLEY_WORKDIR = os.environ['FINLEY_WORKDIR']
except KeyError:
    FINLEY_WORKDIR = '.'

# Determine optimal domain subdivisions for current MPI configuration
mpiSize = getMPISizeWorld()

# For 2D domains
for x in [int(mpiSize**0.5), 2, 3, 5, 7, 1]:
    NX = x
    NY = mpiSize // x
    if NX * NY == mpiSize:
        break

# For 3D domains
for x in [(int(mpiSize**(1/3.)), int(mpiSize**(1/3.))), (2, 3), (2, 2), (1, 2), (1, 1)]:
    NXb = x[0]
    NYb = x[1]
    NZb = mpiSize // (x[0] * x[1])
    if NXb * NYb * NZb == mpiSize:
        break


class Test_MPI_Comm_Finley2D(Test_MPI_Comm_Retrieval):
    """Test MPI communicator retrieval on 2D Finley Rectangle domains"""

    def createDomain(self, comm=None):
        """Create a 2D Rectangle domain"""
        if comm is None:
            return Rectangle(n0=20, n1=20)
        else:
            return Rectangle(n0=20, n1=20, comm=comm)


class Test_MPI_Comm_Finley3D(Test_MPI_Comm_Retrieval):
    """Test MPI communicator retrieval on 3D Finley Brick domains"""

    def createDomain(self, comm=None):
        """Create a 3D Brick domain"""
        if comm is None:
            return Brick(n0=12, n1=12, n2=12)
        else:
            return Brick(n0=12, n1=12, n2=12, comm=comm)


class Test_MPI_Comm_ReadMesh(Test_MPI_Comm_Retrieval):
    """Test MPI communicator retrieval on domains loaded via ReadMesh"""

    def createDomain(self, comm=None):
        """Create a domain by reading a mesh file"""
        mesh_path = os.path.join(FINLEY_TEST_MESH_PATH, 'rect_4x4.fly')
        if comm is None:
            return ReadMesh(mesh_path)
        else:
            return ReadMesh(mesh_path, comm=comm)


class Test_MPI_Comm_ReadGmsh(Test_MPI_Comm_Retrieval):
    """Test MPI communicator retrieval on domains loaded via ReadGmsh"""

    def createDomain(self, comm=None):
        """Create a domain by reading a Gmsh file"""
        mesh_path = os.path.join(FINLEY_TEST_MESH_PATH, 'tagtest.msh')
        if comm is None:
            return ReadGmsh(mesh_path, 2)
        else:
            return ReadGmsh(mesh_path, 2, comm=comm)


class Test_MPI_Comm_LoadMesh(Test_MPI_Comm_Retrieval):
    """Test MPI communicator retrieval on domains loaded via LoadMesh"""

    def createDomain(self, comm=None):
        """Create a domain by loading an HDF5 dump file"""
        from esys.escript import getMPIRankWorld, getMPISizeWorld

        # Use unique filename per MPI world rank to avoid file conflicts
        # when split communicators are used
        world_rank = getMPIRankWorld()
        dump_file = os.path.join(FINLEY_WORKDIR, f'test_load_comm_{world_rank}.dump.h5')

        # Step 1: Create a Rectangle domain with the provided communicator
        if comm is None:
            test_domain = Rectangle(n0=8, n1=8)
        else:
            test_domain = Rectangle(n0=8, n1=8, comm=comm)

        # Step 2: Dump the domain to file
        test_domain.dump(dump_file)

        # Step 3: Verify the dump file exists for this rank
        # Determine which rank we are in the communicator
        if comm is not None:
            my_rank = comm.Get_rank()
            comm_size = comm.Get_size()
        else:
            my_rank = getMPIRankWorld()
            comm_size = getMPISizeWorld()

        # Check if the expected file exists
        # dump() appends rank to filename if comm size > 1: filename.0000, filename.0001, etc.
        if comm_size > 1:
            my_dump_file = f"{dump_file}.{my_rank:04d}"
        else:
            my_dump_file = dump_file

        if not os.path.exists(my_dump_file):
            raise RuntimeError(f"Expected dump file {my_dump_file} not found after dump()")

        # Step 4: Load the domain and return it
        if comm is None:
            return LoadMesh(dump_file)
        else:
            return LoadMesh(dump_file, comm=comm)


if __name__ == '__main__':
    run_tests(__name__, exit_on_failure=True)




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

# =============================================================================
# WINDOWS BUILD OPTIONS - DEPRECATED
# =============================================================================
#
# RECOMMENDED APPROACH: Use WSL2 (Windows Subsystem for Linux)
# -----------------------------------------------------------------------------
# Native Windows builds are complex and not actively tested. The recommended
# approach for Windows users is to use WSL2 with Ubuntu or Debian:
#
#   1. Install WSL2: https://docs.microsoft.com/en-us/windows/wsl/install
#   2. Install Ubuntu from the Microsoft Store
#   3. Inside WSL2, follow the standard Linux installation instructions
#   4. Use debian_options.py or ubuntu_options.py as your options file
#
# This provides a fully supported Linux environment on Windows with much
# simpler dependency management.
#
# =============================================================================
# NATIVE WINDOWS BUILD (unsupported, for reference only)
# =============================================================================
# The settings below are provided for reference only. Native Windows builds
# require manual configuration of all dependencies and are not tested.
#
# Copy this file to <hostname>_options.py, where <hostname> is your machine's
# short hostname, then customize to your needs.
#
# PREFIXES:
# There are two ways to specify where to find dependent headers and libraries:
# 1) If your installation follows the general scheme where headers are located
#    in <prefix>/include, and libraries in <prefix>/lib then it is sufficient
#    to specify this prefix, e.g. boost_prefix='C:/python'
# 2) Otherwise provide a list with two elements, where the first one is the
#    include path, and the second the library path, e.g.
#    boost_prefix=['C:/boost/include/boost1_84', 'C:/boost/lib']

# The options file version. SCons will refuse to build if there have been
# changes to the set of variables and your file has not been updated.
escript_opts_version = 203

# Installation prefix. Files will be installed in subdirectories underneath.
#prefix = 'C:/escript'

# C++ compiler command name or full path.
#cxx = 'cl'

# Additional flags to add to the C++ compiler
#cxx_extra = ''

# Additional flags to add to the linker
#ld_extra = ''

# Whether to treat compiler warnings as errors
#werror = False

# Whether to build a debug version
#debug = False

# Set to True to add flags that enable OpenMP parallelization
#openmp = True

# Flavour of MPI implementation
# Recognized values: 'none', 'MPICH', 'OPENMPI', 'INTELMPI', 'MPT'
#mpi = 'none'

# Prefix or paths to MPI headers and libraries
#mpi_prefix = 'C:/Program Files/Microsoft MPI'

# Prefix or paths to boost-python headers and libraries
#boost_prefix = 'C:/boost'

# boost-python library/libraries to link against
#boost_libs = ['boost_python312']

# Whether to use the netCDF library for dump file support
#netcdf = False

# Prefix or paths to netCDF headers and libraries
#netcdf_prefix = 'C:/netcdf'

# Whether to use UMFPACK
#umfpack = False

# Prefix or paths to UMFPACK headers and libraries
#umfpack_prefix = 'C:/SuiteSparse'

# Flavour of LAPACK implementation
# Recognized values: 'none', 'clapack', 'mkl'
#lapack = 'none'

# Whether to use LLNL's SILO library for output file support
#silo = False

# List of domain families to build
#domains = 'finley,ripley,speckley'

# Extra libraries to link with
#sys_libs = ['ws2_32']

# Tools for SCons
#tools_names = ['msvc']



##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
##############################################################################

"""
Test suite for MPI communicator retrieval (getMPIComm) functionality
"""

__copyright__ = """Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__ = """Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""

import esys.escriptcore.utestselect as unittest
from esys.escriptcore.testing import *
from esys.escript import *

# Check if escript was built with mpi4py support
ESCRIPT_HAS_MPI4PY = haveMPI4Py()

# Check if mpi4py Python module is available AND escript has mpi4py support
try:
    from mpi4py import MPI
    HAVE_MPI4PY = ESCRIPT_HAS_MPI4PY  # Only true if both module AND escript support exist
except ImportError:
    HAVE_MPI4PY = False
    MPI = None


class Test_MPI_Comm_Retrieval(unittest.TestCase):
    """
    Base test class for MPI communicator retrieval.

    Subclasses must implement createDomain() method that returns a domain.
    Tests verify that getMPIComm() correctly returns the communicator for:
    a) No comm argument (default MPI_COMM_WORLD)
    b) Explicit comm=MPI.COMM_WORLD
    c) Custom split communicator (if multiple processes available)
    """

    def createDomain(self, comm=None):
        """
        Create a domain for testing. Must be implemented by subclasses.

        Args:
            comm: Optional MPI communicator to pass to domain constructor

        Returns:
            Domain object
        """
        raise NotImplementedError("Subclasses must implement createDomain()")

    def setUp(self):
        """Set up test fixtures"""
        self.mpiSize = getMPISizeWorld()
        self.mpiRank = getMPIRankWorld()

    def tearDown(self):
        """Clean up after tests"""
        pass

    @unittest.skipIf(not HAVE_MPI4PY, "mpi4py not available")
    def test_getMPIComm_returns_mpi4py_comm(self):
        """Test that getMPIComm() returns an mpi4py.MPI.Comm object"""
        domain = self.createDomain()
        comm = domain.getMPIComm()

        # Should return mpi4py.MPI.Comm object or None
        self.assertTrue(comm is None or isinstance(comm, MPI.Comm),
                        f"Expected MPI.Comm or None, got {type(comm)}")

        # If MPI is enabled, should not be None
        if self.mpiSize > 0:
            self.assertIsNotNone(comm, "getMPIComm() should not return None when MPI is enabled")

    @unittest.skipIf(not HAVE_MPI4PY, "mpi4py not available")
    def test_getMPIComm_no_arg_uses_comm_world(self):
        """Test case (a): No comm argument defaults to MPI_COMM_WORLD"""
        # Create domain without comm argument
        domain = self.createDomain()
        comm = domain.getMPIComm()

        if comm is not None:
            # Should be IDENT or CONGRUENT to MPI_COMM_WORLD
            result = MPI.Comm.Compare(MPI.COMM_WORLD, comm)
            self.assertIn(result, [MPI.IDENT, MPI.CONGRUENT],
                         f"Expected IDENT or CONGRUENT to MPI_COMM_WORLD, got {result}")

            # Size and rank should match
            self.assertEqual(comm.Get_size(), self.mpiSize,
                           "Communicator size mismatch")
            self.assertEqual(comm.Get_rank(), self.mpiRank,
                           "Communicator rank mismatch")

    @unittest.skipIf(not HAVE_MPI4PY, "mpi4py not available")
    def test_getMPIComm_explicit_comm_world(self):
        """Test case (b): Explicit comm=MPI.COMM_WORLD"""
        # Create domain with explicit MPI_COMM_WORLD
        domain = self.createDomain(comm=MPI.COMM_WORLD)
        comm = domain.getMPIComm()

        if comm is not None:
            # Should be IDENT or CONGRUENT to MPI_COMM_WORLD
            result = MPI.Comm.Compare(MPI.COMM_WORLD, comm)
            self.assertIn(result, [MPI.IDENT, MPI.CONGRUENT],
                         f"Expected IDENT or CONGRUENT to MPI_COMM_WORLD, got {result}")

            # Size and rank should match
            self.assertEqual(comm.Get_size(), self.mpiSize,
                           "Communicator size mismatch")
            self.assertEqual(comm.Get_rank(), self.mpiRank,
                           "Communicator rank mismatch")

    @unittest.skipIf(not HAVE_MPI4PY, "mpi4py not available")
    @unittest.skipIf(getMPISizeWorld() < 2, "Requires at least 2 MPI processes")
    def test_getMPIComm_split_comm_size2(self):
        """Test case (c): Custom split communicator with size 2"""
        # Split into groups of size 2 (or as close as possible)
        color = self.mpiRank // 2
        split_comm = MPI.COMM_WORLD.Split(color, self.mpiRank)

        try:
            # Create domain with split communicator
            domain = self.createDomain(comm=split_comm)
            retrieved_comm = domain.getMPIComm()

            if retrieved_comm is not None:
                # Should be IDENT or CONGRUENT to split_comm
                result = MPI.Comm.Compare(split_comm, retrieved_comm)
                self.assertIn(result, [MPI.IDENT, MPI.CONGRUENT],
                             f"Expected IDENT or CONGRUENT to split communicator, got {result}")

                # Size should match (2 or 1 for odd number of processes)
                expected_size = min(2, self.mpiSize - color * 2)
                self.assertEqual(retrieved_comm.Get_size(), expected_size,
                               f"Expected size {expected_size}, got {retrieved_comm.Get_size()}")

                # Rank should be valid within the split communicator
                self.assertLess(retrieved_comm.Get_rank(), retrieved_comm.Get_size(),
                              "Rank should be less than size")
        finally:
            split_comm.Free()

    @unittest.skipIf(not HAVE_MPI4PY, "mpi4py not available")
    @unittest.skipIf(getMPISizeWorld() < 4, "Requires at least 4 MPI processes")
    def test_getMPIComm_split_comm_size4(self):
        """Test case (c): Custom split communicator with size 4"""
        # Split into groups of size 4 (or as close as possible)
        color = self.mpiRank // 4
        split_comm = MPI.COMM_WORLD.Split(color, self.mpiRank)

        try:
            # Create domain with split communicator
            domain = self.createDomain(comm=split_comm)
            retrieved_comm = domain.getMPIComm()

            if retrieved_comm is not None:
                # Should be IDENT or CONGRUENT to split_comm
                result = MPI.Comm.Compare(split_comm, retrieved_comm)
                self.assertIn(result, [MPI.IDENT, MPI.CONGRUENT],
                             f"Expected IDENT or CONGRUENT to split communicator, got {result}")

                # Size should match (4 or less for non-divisible process counts)
                expected_size = min(4, self.mpiSize - color * 4)
                self.assertEqual(retrieved_comm.Get_size(), expected_size,
                               f"Expected size {expected_size}, got {retrieved_comm.Get_size()}")
        finally:
            split_comm.Free()

    @unittest.skipIf(not HAVE_MPI4PY, "mpi4py not available")
    def test_functionspace_getMPIComm(self):
        """Test that FunctionSpace.getMPIComm() returns domain's communicator"""
        domain = self.createDomain()
        domain_comm = domain.getMPIComm()

        # Create function space
        fs = ContinuousFunction(domain)
        fs_comm = fs.getMPIComm()

        if domain_comm is not None and fs_comm is not None:
            # Should be same communicator
            result = MPI.Comm.Compare(domain_comm, fs_comm)
            self.assertIn(result, [MPI.IDENT, MPI.CONGRUENT],
                         "FunctionSpace communicator should match domain communicator")

    @unittest.skipIf(not HAVE_MPI4PY, "mpi4py not available")
    def test_data_getMPIComm(self):
        """Test that Data.getMPIComm() returns domain's communicator"""
        domain = self.createDomain()
        domain_comm = domain.getMPIComm()

        # Create function space and data
        fs = ContinuousFunction(domain)
        data = Scalar(1.0, fs)
        data_comm = data.getMPIComm()

        if domain_comm is not None and data_comm is not None:
            # Should be same communicator
            result = MPI.Comm.Compare(domain_comm, data_comm)
            self.assertIn(result, [MPI.IDENT, MPI.CONGRUENT],
                         "Data communicator should match domain communicator")

    @unittest.skipIf(not HAVE_MPI4PY, "mpi4py not available")
    @unittest.skipIf(getMPISizeWorld() < 2, "Requires at least 2 MPI processes")
    def test_split_comm_consistency_across_objects(self):
        """Test that split communicator is consistent across Domain, FunctionSpace, and Data"""
        # Create split communicator
        color = self.mpiRank % 2
        split_comm = MPI.COMM_WORLD.Split(color, self.mpiRank)

        try:
            # Create domain, function space, and data
            domain = self.createDomain(comm=split_comm)
            fs = ContinuousFunction(domain)
            data = Scalar(1.0, fs)

            # Get communicators
            domain_comm = domain.getMPIComm()
            fs_comm = fs.getMPIComm()
            data_comm = data.getMPIComm()

            if domain_comm is not None and fs_comm is not None and data_comm is not None:
                # All should be the same communicator
                result1 = MPI.Comm.Compare(domain_comm, fs_comm)
                result2 = MPI.Comm.Compare(fs_comm, data_comm)
                result3 = MPI.Comm.Compare(domain_comm, data_comm)

                self.assertIn(result1, [MPI.IDENT, MPI.CONGRUENT],
                             "Domain and FunctionSpace communicators should match")
                self.assertIn(result2, [MPI.IDENT, MPI.CONGRUENT],
                             "FunctionSpace and Data communicators should match")
                self.assertIn(result3, [MPI.IDENT, MPI.CONGRUENT],
                             "Domain and Data communicators should match")

                # All should match the original split communicator
                result_orig = MPI.Comm.Compare(split_comm, domain_comm)
                self.assertIn(result_orig, [MPI.IDENT, MPI.CONGRUENT],
                             "Retrieved communicator should match original split communicator")
        finally:
            split_comm.Free()


# Helper function for running tests
def get_mpi_comm_tests(domain_factory):
    """
    Create a test class for a specific domain factory function.

    Args:
        domain_factory: Function that creates a domain, signature: domain_factory(comm=None)

    Returns:
        Test class that can be used with unittest
    """
    class DomainSpecificTests(Test_MPI_Comm_Retrieval):
        def createDomain(self, comm=None):
            return domain_factory(comm=comm)

    return DomainSpecificTests




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Open Software License version 3.0
http://www.opensource.org/licenses/osl-3.0.php"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

"""
Test suite for PDE solvers on ripley
"""

from test_simplesolve import SimpleSolveTestCase
import esys.escriptcore.utestselect as unittest
from esys.escriptcore.testing import *

from esys.escript import getMPISizeWorld, hasFeature, sqrt
from esys.escript import SolverFramework
from esys.ripley import MultiResolutionDomain
from esys.escript.linearPDEs import SolverOptions

SOLVER="umfpack"
HAVE_REQUESTED_SOLVER = hasFeature(SOLVER)

# number of elements in the spatial directions
NE0=10
NE1=10
NE2=9
mpiSize=getMPISizeWorld()
for x in [int(sqrt(mpiSize)),2,3,5,7,1]:
    NX=x
    NY=mpiSize//x
    if NX*NY == mpiSize:
        break

for x in [(int(mpiSize**(1/3.)),int(mpiSize**(1/3.))),(2,3),(2,2),(1,2),(1,1)]:
    NXb=x[0]
    NYb=x[1]
    NZb=mpiSize//(x[0]*x[1])
    if NXb*NYb*NZb == mpiSize:
        break

def Rectangle(**kwargs):
    m = MultiResolutionDomain(2, **kwargs)
    return m.getLevel(1)

def Brick(**kwargs):
    m = MultiResolutionDomain(3, **kwargs)
    return m.getLevel(1)
    


@unittest.skipIf(not HAVE_REQUESTED_SOLVER, "%s not available"%SOLVER)
@unittest.skipIf(mpiSize > 1, "UMFPACK runs on single rank only.")
class Test_SimpleSolveMultiRes2D_UMFPACK(SimpleSolveTestCase):
    def setUp(self):
        self.domain = Rectangle(n0=NE0*NX-1, n1=NE1*NY-1, d0=NX, d1=NY, framework=SolverFramework.paso())
        self.method = SolverOptions.DIRECT
        self.package = SolverOptions.UMFPACK

    def tearDown(self):
        del self.domain
        
@unittest.skipIf(not HAVE_REQUESTED_SOLVER, "%s not available"%SOLVER)
@unittest.skipIf(mpiSize > 1, "UMFPACK runs on single rank only.")
class Test_SimpleSolveMultiRes3D_UMFPACK(SimpleSolveTestCase):
    def setUp(self):
        self.domain = Brick(n0=NE0*NXb-1, n1=NE1*NYb-1, n2=NE2*NZb-1, d0=NXb, d1=NYb, d2=NZb, framework=SolverFramework.paso())
        self.method = SolverOptions.DIRECT
        self.package = SolverOptions.UMFPACK

    def tearDown(self):
        del self.domain




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################

# No-MPI configuration for Arch Linux.
# Imports the full arch_options.py and disables MPI.
#
# Prerequisites (MPI packages not required):
#   sudo pacman -Sy python python-numpy python-scipy python-sympy python-matplotlib
#   sudo pacman -Sy gcc scons cmake
#   sudo pacman -Sy boost boost-libs suitesparse hdf5 netcdf lapack zlib metis

from scons.templates.arch_options import *

# Disable MPI
mpi = 'none'
mpi4py = False




##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# See CREDITS file for contributors and development history
#
##############################################################################


__copyright__="""Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""
__url__="https://github.com/LutzGross/esys-escript.github.io"

import esys.escriptcore.utestselect as unittest
from esys.escriptcore.testing import *
from test_util import Test_util, Test_Util_SpatialFunctions, \
        Test_Util_SpatialFunctions_noGradOnBoundary, \
        Test_Util_SpatialFunctions_noGradOnBoundary_noContact
from test_types import Test_addition_types
from test_util_interpolation import Test_Util_Interpolation_Dirac
from test_util_integrals import Test_Util_Integration_Dirac
from test_util_NaN_funcs import Test_util_NaN_funcs
from test_util_base import Test_util_base
# from test_copyWithData import Test_copyWithMask

from esys.escript import FunctionOnBoundary, getMPISizeWorld, HAVE_SYMBOLS
from esys.finley import Rectangle, Brick, JoinFaces, ReadMesh
import os

if HAVE_SYMBOLS:
    from test_symfuncs import Test_symfuncs
else:
    @unittest.skip("Skipping symbolic tests since sympy is not available")
    class Test_symfuncs:
        pass

try:
     ESCRIPT_WORKDIR=os.environ['ESCRIPT_WORKDIR']
except KeyError:
     ESCRIPT_WORKDIR='.'

try:
     FINLEY_TEST_DATA=os.environ['FINLEY_TEST_DATA']
except KeyError:
     FINLEY_TEST_DATA='.'

FINLEY_TEST_MESH_PATH=os.path.join(FINLEY_TEST_DATA,"data_meshes")

FINLEY_MERGE_ERROR = "merge: more than 1 processor is not supported yet."

NE=4 # number elements, must be even

#TODO: What is this testing? Is this still needed?
class Test_Data_Addition(Test_addition_types):
    def setUp(self):
        self.domain = Rectangle(NE, NE+1, 2)
        self.workdir=ESCRIPT_WORKDIR

        self.functionspace = FunctionOnBoundary(self.domain)
    def tearDown(self):
       del self.domain


class Test_UtilOnFinley(Test_util,Test_util_NaN_funcs):
   def setUp(self):
       self.workdir = ESCRIPT_WORKDIR
       self.domain = Rectangle(NE, NE + 1, 2)
       self.functionspace = FunctionOnBoundary(self.domain) # due to a bug in escript python needs to hold a reference to the domain
   def tearDown(self):
       del self.functionspace
       del self.domain

# class Test_SymFuncsOnFinley(Test_symfuncs):
#    def setUp(self):
#        try:
#            self.workdir=os.environ['FINLEY_WORKDIR']
#        except KeyError:
#            self.workdir='.'
#        self.domain = Rectangle(NE, NE+1, 2)
#        self.functionspace = FunctionOnBoundary(self.domain) # due to a bug in escript python needs to hold a reference to the domain
#    def tearDown(self):
#        del self.functionspace
#        del self.domain

class Test_NaNFuncsOnFinley(Test_util_NaN_funcs):
   def setUp(self):
       self.workdir = ESCRIPT_WORKDIR
       self.domain = Rectangle(NE, NE+1, 2)
       self.functionspace = FunctionOnBoundary(self.domain) # due to a bug in escript python needs to hold a reference to the domain
   def tearDown(self):
       del self.functionspace
       del self.domain

class Test_Util_SpatialFunctionsOnFinleyTet2DOrder1(Test_Util_SpatialFunctions_noGradOnBoundary_noContact):
    def setUp(self):
        self.order=1
        self.domain = ReadMesh(os.path.join(FINLEY_TEST_MESH_PATH,"tet_2D_order1.fly"),optimize=False)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyTet2DOrder2(Test_Util_SpatialFunctions_noGradOnBoundary_noContact):
    def setUp(self):
        self.order=2
        self.domain = ReadMesh(os.path.join(FINLEY_TEST_MESH_PATH,"tet_2D_order2.fly"),optimize=False)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyTet2DMacro(Test_Util_SpatialFunctions_noGradOnBoundary_noContact):
    def setUp(self):
        self.order=1
        self.domain = ReadMesh(os.path.join(FINLEY_TEST_MESH_PATH,"tet_2D_macro.fly"),optimize=False)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyTet3DOrder1(Test_Util_SpatialFunctions_noGradOnBoundary_noContact):
    def setUp(self):
        self.order=1
        self.domain = ReadMesh(os.path.join(FINLEY_TEST_MESH_PATH,"tet_3D_order1.fly"),optimize=False)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyTet3DOrder2(Test_Util_SpatialFunctions_noGradOnBoundary_noContact):
    def setUp(self):
        self.order=2
        self.domain = ReadMesh(os.path.join(FINLEY_TEST_MESH_PATH,"tet_3D_order2.fly"),optimize=False)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyTet3DMacro(Test_Util_SpatialFunctions_noGradOnBoundary_noContact):
    def setUp(self):
        self.order=1
        self.domain = ReadMesh(os.path.join(FINLEY_TEST_MESH_PATH,"tet_3D_macro.fly"),optimize=False)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyHex2DOrder1(Test_Util_SpatialFunctions_noGradOnBoundary_noContact):
    def setUp(self):
        self.order=1
        self.domain = Rectangle(n0=NE,n1=NE,order=1,useElementsOnFace=0)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyHex2DOrder2(Test_Util_SpatialFunctions_noGradOnBoundary_noContact):
    def setUp(self):
        self.order=2
        self.domain = Rectangle(n0=NE,n1=NE,order=2,useElementsOnFace=0)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyHex2DMacro(Test_Util_SpatialFunctions_noGradOnBoundary_noContact):
    def setUp(self):
        self.order=1
        self.domain = Rectangle(n0=NE,n1=NE,order=-1,useElementsOnFace=0)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyHex3DOrder1(Test_Util_SpatialFunctions_noGradOnBoundary_noContact):
    def setUp(self):
        self.order=1
        self.domain = Brick(n0=NE,n1=NE,n2=NE,order=1,useElementsOnFace=0)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyHex3DOrder2(Test_Util_SpatialFunctions_noGradOnBoundary_noContact):
    def setUp(self):
        self.order=2
        self.domain = Brick(n0=NE,n1=NE,n2=NE,order=2,useElementsOnFace=0)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyHex3DMacro(Test_Util_SpatialFunctions_noGradOnBoundary_noContact):
    def setUp(self):
        self.order=1
        self.domain = Brick(n0=NE,n1=NE,n2=NE,order=-1,useElementsOnFace=0)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyHex2DOrder1withContact(Test_Util_SpatialFunctions_noGradOnBoundary):
    def setUp(self):
        self.order=1
        d1 = Rectangle(n0=NE//2+1,n1=NE,l0=0.5,order=1,useElementsOnFace=0)
        d2 = Rectangle(n0=NE//2,n1=NE,l0=0.5,order=1,useElementsOnFace=0)
        d2.setX(d2.getX()+[0.5,0.])
        if getMPISizeWorld() > 1:
            with self.assertRaises(NotImplementedError) as pkg:
                self.domain = JoinFaces([d1,d2],optimize=False)
            e = pkg.exception
            if FINLEY_MERGE_ERROR not in str(e):
                raise e
            raise unittest.SkipTest(FINLEY_MERGE_ERROR)
        else:
            self.domain = JoinFaces([d1,d2],optimize=False)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyHex2DOrder2withContact(Test_Util_SpatialFunctions_noGradOnBoundary):
    def setUp(self):
        self.order=2
        d1 = Rectangle(n0=NE//2,n1=NE,l0=0.5,order=2,useElementsOnFace=0)
        d2 = Rectangle(n0=NE//2,n1=NE,l0=0.5,order=2,useElementsOnFace=0)
        d2.setX(d2.getX()+[0.5,0.])
        if getMPISizeWorld() > 1:
            with self.assertRaises(NotImplementedError) as pkg:
                self.domain = JoinFaces([d1,d2],optimize=False)
            e = pkg.exception
            if FINLEY_MERGE_ERROR not in str(e):
                raise e
            raise unittest.SkipTest(FINLEY_MERGE_ERROR)
        else:
            self.domain = JoinFaces([d1,d2],optimize=False)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyHex3DOrder1withContact(Test_Util_SpatialFunctions_noGradOnBoundary):
    def setUp(self):
        self.order=1
        d1 = Brick(n0=NE//2+1,n1=NE,n2=NE,l0=0.5,order=1,useElementsOnFace=0)
        d2 = Brick(n0=NE//2,n1=NE,n2=NE,l0=0.5,order=1,useElementsOnFace=0)
        d2.setX(d2.getX()+[0.5,0.,0.])
        if getMPISizeWorld() > 1:
            with self.assertRaises(NotImplementedError) as pkg:
                self.domain = JoinFaces([d1,d2],optimize=False)
            e = pkg.exception
            if FINLEY_MERGE_ERROR not in str(e):
                raise e
            raise unittest.SkipTest(FINLEY_MERGE_ERROR)
        else:
            self.domain = JoinFaces([d1,d2],optimize=False)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyHex3DOrder2withContact(Test_Util_SpatialFunctions_noGradOnBoundary):
    def setUp(self):
        self.order=2
        d1 = Brick(n0=NE//2+1,n1=NE,n2=NE,l0=0.5,order=2,useElementsOnFace=0)
        d2 = Brick(n0=NE//2,n1=NE,n2=NE,l0=0.5,order=2,useElementsOnFace=0)
        d2.setX(d2.getX()+[0.5,0.,0.])
        if getMPISizeWorld() > 1:
            with self.assertRaises(NotImplementedError) as pkg:
                self.domain = JoinFaces([d1,d2],optimize=False)
            e = pkg.exception
            if FINLEY_MERGE_ERROR not in str(e):
                raise e
            raise unittest.SkipTest(FINLEY_MERGE_ERROR)
        else:
            self.domain = JoinFaces([d1,d2],optimize=False)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyHex2DOrder1useElementsOnFacewithContact(Test_Util_SpatialFunctions):
    def setUp(self):
        self.order=1
        d1 = Rectangle(n0=NE//2+1,n1=NE,l0=0.5,order=1,useElementsOnFace=True)
        d2 = Rectangle(n0=NE//2,n1=NE,l0=0.5,order=1,useElementsOnFace=True)
        d2.setX(d2.getX()+[0.5,0.])
        if getMPISizeWorld() > 1:
            with self.assertRaises(NotImplementedError) as pkg:
                self.domain = JoinFaces([d1,d2],optimize=False)
            e = pkg.exception
            if FINLEY_MERGE_ERROR not in str(e):
                raise e
            raise unittest.SkipTest(FINLEY_MERGE_ERROR)
        else:
            self.domain = JoinFaces([d1,d2],optimize=False)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyHex2DOrder2useElementsOnFacewithContact(Test_Util_SpatialFunctions):
    def setUp(self):
        self.order=2
        d1 = Rectangle(n0=NE//2+1,n1=NE,l0=0.5,order=2,useElementsOnFace=True)
        d2 = Rectangle(n0=NE//2,n1=NE,l0=0.5,order=2,useElementsOnFace=True)
        d2.setX(d2.getX()+[0.5,0.])
        if getMPISizeWorld() > 1:
            with self.assertRaises(NotImplementedError) as pkg:
                self.domain = JoinFaces([d1,d2],optimize=False)
            e = pkg.exception
            if FINLEY_MERGE_ERROR not in str(e):
                raise e
            raise unittest.SkipTest(FINLEY_MERGE_ERROR)
        else:
            self.domain = JoinFaces([d1,d2],optimize=False)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyHex3DOrder1useElementsOnFacewithContact(Test_Util_SpatialFunctions):
    def setUp(self):
        self.order=1
        d1 = Brick(n0=NE//2,n1=NE,n2=NE,l0=0.5,order=1,useElementsOnFace=True)
        d2 = Brick(n0=NE//2+1,n1=NE,n2=NE,l0=0.5,order=1,useElementsOnFace=True)
        d2.setX(d2.getX()+[0.5,0.,0.])
        if getMPISizeWorld() > 1:
            with self.assertRaises(NotImplementedError) as pkg:
                self.domain = JoinFaces([d1,d2],optimize=False)
            e = pkg.exception
            if FINLEY_MERGE_ERROR not in str(e):
                raise e
            raise unittest.SkipTest(FINLEY_MERGE_ERROR)
        else:
            self.domain = JoinFaces([d1,d2],optimize=False)
    def tearDown(self):
        del self.order
        del self.domain

class Test_Util_SpatialFunctionsOnFinleyHex3DOrder2useElementsOnFacewithContact(Test_Util_SpatialFunctions):
    def setUp(self):
        self.order=2
        d1 = Brick(n0=NE//2,n1=NE,n2=NE,l0=0.5,order=2,useElementsOnFace=True)
        d2 = Brick(n0=NE//2+1,n1=NE,n2=NE,l0=0.5,order=2,useElementsOnFace=True)
        d2.setX(d2.getX()+[0.5,0.,0.])
        if getMPISizeWorld() > 1:
            with self.assertRaises(NotImplementedError) as pkg:
                self.domain = JoinFaces([d1,d2],optimize=False)
            e = pkg.exception
            if FINLEY_MERGE_ERROR not in str(e):
                raise e
            raise unittest.SkipTest(FINLEY_MERGE_ERROR)
        else:
            self.domain = JoinFaces([d1,d2],optimize=False)
    def tearDown(self):
        del self.order
        del self.domain

class Test_2D_Point_Order1_Interpolation(Test_Util_Interpolation_Dirac):
    def setUp(self):
        Stations = [ (0.,0.), (1.,0), (0,1), (1,1) ]
        StationsTags = ["A1", "A2", "A3", "A4" ]
        self.positions=Stations
        self.taglist=StationsTags
        self.domain=Rectangle(n0=5,n1=5, diracPoints=Stations, diracTags=StationsTags)
    def tearDown(self):
        del self.domain, self.positions, self.taglist

class Test_3D_Point_Order1_Interpolation(Test_Util_Interpolation_Dirac):
    def setUp(self):
        Stations = [ (0.,0.,0.), (1.,0,0.), (0,1,0.), (1,1,0.), (0.,0.,1.), (1.,0,1.), (0,1,1.), (1,1,1.) ]
        StationsTags = ["A1", "A2", "A3", "A4","A5", "A6", "A7", "A8"  ]
        self.positions=Stations
        self.taglist=StationsTags
        self.domain=Brick(n0=5,n1=5,n2=5,diracPoints=Stations,diracTags=StationsTags)
    def tearDown(self):
        del self.domain, self.positions, self.taglist

class Test_2D_Point_Order2_Interpolation(Test_Util_Interpolation_Dirac):
    def setUp(self):
        Stations = [ (0.,0.), (1.,0), (0,1), (1,1) ]
        StationsTags = ["A1", "A2", "A3", "A4" ]
        self.positions=Stations
        self.taglist=StationsTags
        self.domain=Rectangle(n0=5,n1=5, order=2, diracPoints=Stations, diracTags=StationsTags)
    def tearDown(self):
        del self.domain, self.positions, self.taglist

class Test_3D_Point_Order2_Interpolation(Test_Util_Interpolation_Dirac):
    def setUp(self):
        Stations = [ (0.,0.,0.), (1.,0,0.), (0,1,0.), (1,1,0.), (0.,0.,1.), (1.,0,1.), (0,1,1.), (1,1,1.) ]
        StationsTags = ["A1", "A2", "A3", "A4","A5", "A6", "A7", "A8"  ]
        self.positions=Stations
        self.taglist=StationsTags
        self.domain=Brick(n0=5,n1=5,n2=5,order=2, diracPoints=Stations,diracTags=StationsTags)
    def tearDown(self):
        del self.domain, self.positions, self.taglist

class Test_2D_Point_Order1_Integration(Test_Util_Integration_Dirac):
    def setUp(self):
        Stations = [ (0.,0.), (1.,0), (0,1), (1,1) ]
        StationsTags = ["A1", "A2", "A3", "A4" ]
        self.taglist=StationsTags
        self.domain=Rectangle(n0=5,n1=5, diracPoints=Stations, diracTags=StationsTags)
    def tearDown(self):
        del self.domain, self.taglist

class Test_3D_Point_Order1_Integration(Test_Util_Integration_Dirac):
    def setUp(self):
        Stations = [ (0.,0.,0.), (1.,0,0.), (0,1,0.), (1,1,0.), (0.,0.,1.), (1.,0,1.), (0,1,1.), (1,1,1.) ]
        StationsTags = ["A1", "A2", "A3", "A4","A5", "A6", "A7", "A8"  ]
        self.taglist=StationsTags
        self.domain=Brick(n0=5,n1=5,n2=5,diracPoints=Stations,diracTags=StationsTags)
    def tearDown(self):
        del self.domain, self.taglist
class Test_2D_Point_Order2_Integration(Test_Util_Integration_Dirac):
    def setUp(self):
        Stations = [ (0.,0.), (1.,0), (0,1), (1,1) ]
        StationsTags = ["A1", "A2", "A3", "A4" ]
        self.taglist=StationsTags
        self.domain=Rectangle(n0=5,n1=5, order=2, diracPoints=Stations, diracTags=StationsTags)
    def tearDown(self):
        del self.domain, self.taglist

class Test_3D_Point_Order2_Integration(Test_Util_Integration_Dirac):
    def setUp(self):
        Stations = [ (0.,0.,0.), (1.,0,0.), (0,1,0.), (1,1,0.), (0.,0.,1.), (1.,0,1.), (0,1,1.), (1,1,1.) ]
        StationsTags = ["A1", "A2", "A3", "A4","A5", "A6", "A7", "A8"  ]
        self.taglist=StationsTags
        self.domain=Brick(n0=5,n1=5,n2=5,order=2, diracPoints=Stations,diracTags=StationsTags)
    def tearDown(self):
        del self.domain, self.taglist

class Test_DiameterOnFinley2D(Test_util_base):
    def setUp(self):
        self.workdir = ESCRIPT_WORKDIR
        self.domain = Rectangle(NE, NE+1, 2)
        self.functionspace = FunctionOnBoundary(self.domain)
    def tearDown(self):
        del self.functionspace
        del self.domain

class Test_DiameterOnFinley3D(Test_util_base):
    def setUp(self):
        self.workdir = ESCRIPT_WORKDIR
        self.domain = Brick(NE, NE, NE)
        self.functionspace = FunctionOnBoundary(self.domain)
    def tearDown(self):
        del self.functionspace
        del self.domain

# TODO
# class Test_copyWithMask_rectangle(Test_copyWithMask):
#     def setUp(self):
#         self.domain=Rectangle(n0=5,n1=7)
#     def tearDown(self):
#         del self.domain

if __name__ == '__main__':
    run_tests(__name__, exit_on_failure=True)



##############################################################################
#
# Copyright (c) 2003-2026 by the esys.escript Group
# https://github.com/LutzGross/esys-escript.github.io
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
##############################################################################

"""
GOCAD Voxet Reader Example

This script demonstrates how to read GOCAD voxet data and use it with
Ripley domains in escript.

GOCAD voxet files store 3D gridded data, typically used in geosciences
for representing subsurface properties like density, velocity, or resistivity.

Note: This is a placeholder example showing the general structure.
The actual voxet reading functionality may require additional libraries
or custom parsing of GOCAD file formats.

Usage:
    run-escript voxet_reader.py <voxet_file>
"""

__copyright__ = """Copyright (c) 2003-2026 by the esys.escript Group
https://github.com/LutzGross/esys-escript.github.io
Primary Business: Queensland, Australia"""
__license__ = """Licensed under the Apache License, version 2.0
http://www.apache.org/licenses/LICENSE-2.0"""

import sys
import numpy as np
from esys.escript import *
from esys.ripley import Brick


def read_gocad_voxet(filename):
    """
    Read a GOCAD voxet file.

    This is a placeholder function. A full implementation would parse
    the GOCAD voxet format which typically consists of:
    - Header file (.vo) with grid geometry and properties
    - Binary data file with the actual property values

    Parameters
    ----------
    filename : str
        Path to the voxet header file

    Returns
    -------
    dict
        Dictionary containing:
        - 'nx', 'ny', 'nz': Number of cells in each direction
        - 'origin': Origin coordinates (x0, y0, z0)
        - 'spacing': Cell spacing (dx, dy, dz)
        - 'data': 3D numpy array of property values
    """
    print(f"Reading GOCAD voxet file: {filename}")
    print("Note: This is a placeholder implementation")

    # Placeholder: In reality, parse the voxet header file
    # and read the binary data file
    voxet_data = {
        'nx': 50,
        'ny': 50,
        'nz': 20,
        'origin': (0.0, 0.0, 0.0),
        'spacing': (100.0, 100.0, 50.0),
        'data': np.random.rand(50, 50, 20)  # Placeholder data
    }

    return voxet_data


def create_ripley_domain_from_voxet(voxet_data):
    """
    Create a Ripley Brick domain matching the voxet geometry.

    Parameters
    ----------
    voxet_data : dict
        Voxet data as returned by read_gocad_voxet()

    Returns
    -------
    Domain
        Ripley Brick domain
    """
    nx = voxet_data['nx']
    ny = voxet_data['ny']
    nz = voxet_data['nz']

    x0, y0, z0 = voxet_data['origin']
    dx, dy, dz = voxet_data['spacing']

    # Calculate domain extents
    l0 = nx * dx
    l1 = ny * dy
    l2 = nz * dz

    # Create Ripley Brick domain
    domain = Brick(n0=nx, n1=ny, n2=nz, l0=l0, l1=l1, l2=l2,
                   d0=1, d1=1, d2=1)  # No MPI subdivision for simplicity

    print(f"Created Ripley domain: {nx}x{ny}x{nz} cells")
    print(f"Domain extent: {l0}x{l1}x{l2}")

    return domain


def map_voxet_data_to_domain(voxet_data, domain, function_space=None):
    """
    Map voxet data values onto a Ripley domain.

    Parameters
    ----------
    voxet_data : dict
        Voxet data as returned by read_gocad_voxet()
    domain : Domain
        Ripley domain
    function_space : FunctionSpace, optional
        Target function space (default: ReducedFunction)

    Returns
    -------
    Data
        escript Data object containing the voxet values
    """
    if function_space is None:
        function_space = ReducedFunction(domain)

    # Flatten the 3D voxet data to 1D array
    # Note: Ordering may need adjustment depending on voxet format
    values_flat = voxet_data['data'].flatten()

    # Create escript Data object
    # This assumes the voxet grid matches the domain grid
    data = Scalar(0.0, function_space)

    # In a real implementation, you would map the voxet values
    # to the appropriate locations in the escript Data object
    # This may require interpolation if grids don't match exactly

    print(f"Mapped voxet data to domain")
    print(f"Data range: [{np.min(values_flat):.3f}, {np.max(values_flat):.3f}]")

    return data


def main():
    """Main function to demonstrate voxet reading workflow."""

    if len(sys.argv) < 2:
        print("Usage: run-escript voxet_reader.py <voxet_file>")
        print("\nNote: This is a placeholder example.")
        print("Using synthetic data for demonstration.")
        voxet_file = "example.vo"
    else:
        voxet_file = sys.argv[1]

    # Read voxet data
    voxet_data = read_gocad_voxet(voxet_file)

    # Create matching Ripley domain
    domain = create_ripley_domain_from_voxet(voxet_data)

    # Map voxet data onto domain
    property_data = map_voxet_data_to_domain(voxet_data, domain)

    # Example: Use the data in a PDE
    # (This would be application-specific)
    print("\nVoxet data successfully loaded into Ripley domain")
    print("The property_data can now be used in PDEs or exported")

    # Example export
    # from esys.weipa import saveVTK
    # saveVTK("voxet_output.vtu", property=property_data)


if __name__ == "__main__":
    main()




##############################################################################
#
# Copyright (c) 2003-2020 by The University of Queensland
# http://www.uq.edu.au
#
# Primary Business: Queensland, Australia
# Licensed under the Apache License, version 2.0
# http://www.apache.org/licenses/LICENSE-2.0
#
# Development until 2012 by Earth Systems Science Computational Center (ESSCC)
# Development 2012-2013 by School of Earth Sciences
# Development from 2014 by Centre for Geoscience Computing (GeoComp)
# Development from 2019 by School of Earth and Environmental Sciences
#
##############################################################################
"""SCons.Tool.nvcc

Tool-specific initialization for NVIDIA CUDA Compiler.

There normally shouldn't be any need to import this module directly.
It will usually be imported through the generic SCons.Tool.Tool()
selection method.

"""


import SCons.Tool
import SCons.Scanner.C
import SCons.Defaults
import os
import platform


def get_cuda_paths():
  """Determines CUDA {bin,lib,include} paths
  
  returns (bin_path,lib_path,inc_path)
  """

  # determine defaults
  if os.name == 'nt':
    bin_path = 'C:/CUDA/bin'
    lib_path = 'C:/CUDA/lib'
    inc_path = 'C:/CUDA/include'
  elif os.name == 'posix':
    bin_path = '/usr/local/cuda/bin'
    lib_path = '/usr/local/cuda/lib'
    inc_path = '/usr/local/cuda/include'
  else:
    raise ValueError, 'Error: unknown OS.  Where is nvcc installed?'
   
  if platform.machine()[-2:] == '64':
    lib_path += '64'

  # override with environement variables
  if 'CUDA_BIN_PATH' in os.environ:
    bin_path = os.path.abspath(os.environ['CUDA_BIN_PATH'])
  if 'CUDA_LIB_PATH' in os.environ:
    lib_path = os.path.abspath(os.environ['CUDA_LIB_PATH'])
  if 'CUDA_INC_PATH' in os.environ:
    inc_path = os.path.abspath(os.environ['CUDA_INC_PATH'])

  return (bin_path,lib_path,inc_path)



CUDASuffixes = ['.cu']

# make a CUDAScanner for finding #includes
# cuda uses the c preprocessor, so we can use the CScanner
CUDAScanner = SCons.Scanner.C.CScanner()

def add_common_nvcc_variables(env):
  """
  Add underlying common "NVIDIA CUDA compiler" variables that
  are used by multiple builders.
  """

  # "NVCC common command line"
  if not env.has_key('_NVCCCOMCOM'):
    # nvcc needs '-I' prepended before each include path, regardless of platform
    env['_NVCCWRAPCPPPATH'] = '${_concat("-I ", CPPPATH, "", __env__)}'
    # prepend -Xcompiler before each flag
    env['_NVCCWRAPCFLAGS'] =     '${_concat("-Xcompiler ", CFLAGS,     "", __env__)}'
    env['_NVCCWRAPSHCFLAGS'] =   '${_concat("-Xcompiler ", SHCFLAGS,   "", __env__)}'
    env['_NVCCWRAPCCFLAGS'] =   '${_concat("-Xcompiler ", CCFLAGS,   "", __env__)}'
    env['_NVCCWRAPSHCCFLAGS'] = '${_concat("-Xcompiler ", SHCCFLAGS, "", __env__)}'
    # assemble the common command line
    env['_NVCCCOMCOM'] = '${_concat("-Xcompiler ", CPPFLAGS, "", __env__)} $_CPPDEFFLAGS $_NVCCWRAPCPPPATH'

def generate(env):
  """
  Add Builders and construction variables for CUDA compilers to an Environment.
  """

  # create a builder that makes PTX files from .cu files
  ptx_builder = SCons.Builder.Builder(action = '$NVCC -ptx $NVCCFLAGS $_NVCCWRAPCFLAGS $NVCCWRAPCCFLAGS $_NVCCCOMCOM $SOURCES -o $TARGET',
                                      emitter = {},
                                      suffix = '.ptx',
                                      src_suffix = CUDASuffixes)
  env['BUILDERS']['PTXFile'] = ptx_builder

  # create builders that make static & shared objects from .cu files
  static_obj, shared_obj = SCons.Tool.createObjBuilders(env)

  for suffix in CUDASuffixes:
    # Add this suffix to the list of things buildable by Object
    static_obj.add_action('$CUDAFILESUFFIX', '$NVCCCOM')
    shared_obj.add_action('$CUDAFILESUFFIX', '$SHNVCCCOM')
    static_obj.add_emitter(suffix, SCons.Defaults.StaticObjectEmitter)
    shared_obj.add_emitter(suffix, SCons.Defaults.SharedObjectEmitter)

    # Add this suffix to the list of things scannable
    SCons.Tool.SourceFileScanner.add_scanner(suffix, CUDAScanner)

  add_common_nvcc_variables(env)

  # set the "CUDA Compiler Command" environment variable
  # windows is picky about getting the full filename of the executable
  if os.name == 'nt':
    env['NVCC'] = 'nvcc.exe'
    env['SHNVCC'] = 'nvcc.exe'
  else:
    env['NVCC'] = 'nvcc'
    env['SHNVCC'] = 'nvcc'
  
  # set the include path, and pass both c compiler flags and c++ compiler flags
  env['NVCCFLAGS'] = SCons.Util.CLVar('')
  env['SHNVCCFLAGS'] = SCons.Util.CLVar('') + ' -shared'
  
  # 'NVCC Command'
  env['NVCCCOM']   = '$NVCC -ccbin=g++-4.8 -o $TARGET -c $NVCCFLAGS $_NVCCWRAPCFLAGS $NVCCWRAPCCFLAGS $_NVCCCOMCOM $SOURCES'
  env['SHNVCCCOM'] = '$SHNVCC -ccbin=g++-4.8 -o $TARGET -c $SHNVCCFLAGS $_NVCCWRAPSHCFLAGS $_NVCCWRAPSHCCFLAGS $_NVCCCOMCOM $SOURCES'
  
  # the suffix of CUDA source files is '.cu'
  env['CUDAFILESUFFIX'] = '.cu'

  # XXX add code to generate builders for other miscellaneous
  # CUDA files here, such as .gpu, etc.

  # XXX intelligently detect location of nvcc and cuda libraries here
  (bin_path,lib_path,inc_path) = get_cuda_paths()
    
  env.PrependENVPath('PATH', bin_path)

def exists(env):
  return env.Detect('nvcc')

